#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re, shlex, shutil, sys, time, traceback
from waflib import ConfigSet, Utils, Options, Logs, Context, Build, Errors

WAF_CONFIG_LOG = 'config.log'
autoconfig = False
conf_template = '''# project %(app)s configured on %(now)s by
# waf %(wafver)s (abi %(abi)s, python %(pyver)x on %(systype)s)
# using %(args)s
#'''


class ConfigurationContext(Context.Context):
    '''configures the project'''
    cmd = 'configure'
    error_handlers = []

    def __init__(self, **kw):
        super(ConfigurationContext, self).__init__(**kw)
        self.environ = dict(os.environ)
        self.all_envs = {}
        self.top_dir = None
        self.out_dir = None
        self.tools = []
        self.hash = 0
        self.files = []
        self.tool_cache = []
        self.setenv('')

    def setenv(self, name, env=None):
        if name not in self.all_envs or env:
            if not env:
                env = ConfigSet.ConfigSet()
                self.prepare_env(env)
            else:
                env = env.derive()
            self.all_envs[name] = env
        self.variant = name

    def get_env(self):
        return self.all_envs[self.variant]

    def set_env(self, val):
        self.all_envs[self.variant] = val

    env = property(get_env, set_env)

    def init_dirs(self):
        top = self.top_dir
        if not top:
            top = Options.options.top
        if not top:
            top = getattr(Context.g_module, Context.TOP, None)
        if not top:
            top = self.path.abspath()
        top = os.path.abspath(top)
        self.srcnode = (os.path.isabs(top) and self.root or self.path).find_dir(top)
        assert (self.srcnode)
        out = self.out_dir
        if not out:
            out = Options.options.out
        if not out:
            out = getattr(Context.g_module, Context.OUT, None)
        if not out:
            out = Options.lockfile.replace('.lock-waf_%s_' % sys.platform, '').replace('.lock-waf', '')
        out = os.path.realpath(out)
        self.bldnode = (os.path.isabs(out) and self.root or self.path).make_node(out)
        self.bldnode.mkdir()
        if not os.path.isdir(self.bldnode.abspath()):
            self.fatal('Could not create the build directory %s' % self.bldnode.abspath())

    def execute(self):
        self.init_dirs()
        self.cachedir = self.bldnode.make_node(Build.CACHE_DIR)
        self.cachedir.mkdir()
        path = os.path.join(self.bldnode.abspath(), WAF_CONFIG_LOG)
        self.logger = Logs.make_logger(path, 'cfg')
        app = getattr(Context.g_module, 'APPNAME', '')
        if app:
            ver = getattr(Context.g_module, 'VERSION', '')
            if ver:
                app = "%s (%s)" % (app, ver)
        params = {
            'now': time.ctime(),
            'pyver': sys.hexversion,
            'systype': sys.platform,
            'args': " ".join(sys.argv),
            'wafver': Context.WAFVERSION,
            'abi': Context.ABI,
            'app': app
        }
        self.to_log(conf_template % params)
        self.msg('Setting top to', self.srcnode.abspath())
        self.msg('Setting out to', self.bldnode.abspath())
        if id(self.srcnode) == id(self.bldnode):
            Logs.warn('Setting top == out')
        elif id(self.path) != id(self.srcnode):
            if self.srcnode.is_child_of(self.path):
                Logs.warn('Are you certain that you do not want to set top="." ?')
        super(ConfigurationContext, self).execute()
        self.store()
        Context.top_dir = self.srcnode.abspath()
        Context.out_dir = self.bldnode.abspath()
        env = ConfigSet.ConfigSet()
        env.argv = sys.argv
        env.options = Options.options.__dict__
        env.config_cmd = self.cmd
        env.run_dir = Context.run_dir
        env.top_dir = Context.top_dir
        env.out_dir = Context.out_dir
        env.hash = self.hash
        env.files = self.files
        env.environ = dict(self.environ)
        env.launch_dir = Context.launch_dir
        if not (
            self.env.NO_LOCK_IN_RUN or env.environ.get('NO_LOCK_IN_RUN') or getattr(Options.options, 'no_lock_in_run')
        ):
            env.store(os.path.join(Context.run_dir, Options.lockfile))
        if not (
            self.env.NO_LOCK_IN_TOP or env.environ.get('NO_LOCK_IN_TOP') or getattr(Options.options, 'no_lock_in_top')
        ):
            env.store(os.path.join(Context.top_dir, Options.lockfile))
        if not (
            self.env.NO_LOCK_IN_OUT or env.environ.get('NO_LOCK_IN_OUT') or getattr(Options.options, 'no_lock_in_out')
        ):
            env.store(os.path.join(Context.out_dir, Options.lockfile))

    def prepare_env(self, env):
        if not env.PREFIX:
            if Options.options.prefix or Utils.is_win32:
                env.PREFIX = Options.options.prefix
            else:
                env.PREFIX = '/'
        if not env.BINDIR:
            if Options.options.bindir:
                env.BINDIR = Options.options.bindir
            else:
                env.BINDIR = Utils.subst_vars('${PREFIX}/bin', env)
        if not env.LIBDIR:
            if Options.options.libdir:
                env.LIBDIR = Options.options.libdir
            else:
                env.LIBDIR = Utils.subst_vars('${PREFIX}/lib%s' % Utils.lib64(), env)

    def store(self):
        n = self.cachedir.make_node('build.config.py')
        n.write('version = 0x%x\ntools = %r\n' % (Context.HEXVERSION, self.tools))
        if not self.all_envs:
            self.fatal('nothing to store in the configuration context!')
        for key in self.all_envs:
            tmpenv = self.all_envs[key]
            tmpenv.store(os.path.join(self.cachedir.abspath(), key + Build.CACHE_SUFFIX))

    def load(self, tool_list, tooldir=None, funs=None, with_sys_path=True, cache=False):
        tools = Utils.to_list(tool_list)
        if tooldir:
            tooldir = Utils.to_list(tooldir)
        for tool in tools:
            if cache:
                mag = (tool, id(self.env), tooldir, funs)
                if mag in self.tool_cache:
                    self.to_log('(tool %s is already loaded, skipping)' % tool)
                    continue
                self.tool_cache.append(mag)
            module = None
            try:
                module = Context.load_tool(tool, tooldir, ctx=self, with_sys_path=with_sys_path)
            except ImportError as e:
                self.fatal(
                    'Could not load the Waf tool %r from %r\n%s' % (tool, getattr(e, 'waf_sys_path', sys.path), e)
                )
            except Exception as e:
                self.to_log('imp %r (%r & %r)' % (tool, tooldir, funs))
                self.to_log(traceback.format_exc())
                raise
            if funs is not None:
                self.eval_rules(funs)
            else:
                func = getattr(module, 'configure', None)
                if func:
                    if type(func) is type(Utils.readf):
                        func(self)
                    else:
                        self.eval_rules(func)
            self.tools.append({'tool': tool, 'tooldir': tooldir, 'funs': funs})

    def post_recurse(self, node):
        super(ConfigurationContext, self).post_recurse(node)
        self.hash = Utils.h_list((self.hash, node.read('rb')))
        self.files.append(node.abspath())

    def eval_rules(self, rules):
        self.rules = Utils.to_list(rules)
        for x in self.rules:
            f = getattr(self, x)
            if not f:
                self.fatal('No such configuration function %r' % x)
            f()


def conf(f):
    def fun(*k, **kw):
        mandatory = kw.pop('mandatory', True)
        try:
            return f(*k, **kw)
        except Errors.ConfigurationError:
            if mandatory:
                raise

    fun.__name__ = f.__name__
    setattr(ConfigurationContext, f.__name__, fun)
    setattr(Build.BuildContext, f.__name__, fun)
    return f


@conf
def add_os_flags(self, var, dest=None, dup=False):
    try:
        flags = shlex.split(self.environ[var])
    except KeyError:
        return
    if dup or ''.join(flags) not in ''.join(Utils.to_list(self.env[dest or var])):
        self.env.append_value(dest or var, flags)


@conf
def cmd_to_list(self, cmd):
    if isinstance(cmd, str):
        if os.path.isfile(cmd):
            return [cmd]
        if os.sep == '/':
            return shlex.split(cmd)
        else:
            try:
                return shlex.split(cmd, posix=False)
            except TypeError:
                return shlex.split(cmd)
    return cmd


@conf
def check_waf_version(self, mini='1.9.99', maxi='2.1.0', **kw):
    self.start_msg('Checking for waf version in %s-%s' % (str(mini), str(maxi)), **kw)
    ver = Context.HEXVERSION
    if Utils.num2ver(mini) > ver:
        self.fatal('waf version should be at least %r (%r found)' % (Utils.num2ver(mini), ver))
    if Utils.num2ver(maxi) < ver:
        self.fatal('waf version should be at most %r (%r found)' % (Utils.num2ver(maxi), ver))
    self.end_msg('ok', **kw)


@conf
def find_file(self, filename, path_list=[]):
    for n in Utils.to_list(filename):
        for d in Utils.to_list(path_list):
            p = os.path.expanduser(os.path.join(d, n))
            if os.path.exists(p):
                return p
    self.fatal('Could not find %r' % filename)


@conf
def find_program(self, filename, **kw):
    exts = kw.get('exts', Utils.is_win32 and '.exe,.com,.bat,.cmd' or ',.sh,.pl,.py')
    environ = kw.get('environ', getattr(self, 'environ', os.environ))
    ret = ''
    filename = Utils.to_list(filename)
    msg = kw.get('msg', ', '.join(filename))
    var = kw.get('var', '')
    if not var:
        var = re.sub(r'[-.]', '_', filename[0].upper())
    path_list = kw.get('path_list', '')
    if path_list:
        path_list = Utils.to_list(path_list)
    else:
        path_list = environ.get('PATH', '').split(os.pathsep)
    if kw.get('value'):
        ret = self.cmd_to_list(kw['value'])
    elif environ.get(var):
        ret = self.cmd_to_list(environ[var])
    elif self.env[var]:
        ret = self.cmd_to_list(self.env[var])
    else:
        if not ret:
            ret = self.find_binary(filename, exts.split(','), path_list)
        if not ret and Utils.winreg:
            ret = Utils.get_registry_app_path(Utils.winreg.HKEY_CURRENT_USER, filename)
        if not ret and Utils.winreg:
            ret = Utils.get_registry_app_path(Utils.winreg.HKEY_LOCAL_MACHINE, filename)
        ret = self.cmd_to_list(ret)
    if ret:
        if len(ret) == 1:
            retmsg = ret[0]
        else:
            retmsg = ret
    else:
        retmsg = False
    self.msg('Checking for program %r' % msg, retmsg, **kw)
    if not kw.get('quiet'):
        self.to_log('find program=%r paths=%r var=%r -> %r' % (filename, path_list, var, ret))
    if not ret:
        self.fatal(kw.get('errmsg', '') or 'Could not find the program %r' % filename)
    interpreter = kw.get('interpreter')
    if interpreter is None:
        if not Utils.check_exe(ret[0], env=environ):
            self.fatal('Program %r is not executable' % ret)
        self.env[var] = ret
    else:
        self.env[var] = self.env[interpreter] + ret
    return ret


@conf
def find_binary(self, filenames, exts, paths):
    for f in filenames:
        for ext in exts:
            exe_name = f + ext
            if os.path.isabs(exe_name):
                if os.path.isfile(exe_name):
                    return exe_name
            else:
                for path in paths:
                    x = os.path.expanduser(os.path.join(path, exe_name))
                    if os.path.isfile(x):
                        return x
    return None


@conf
def run_build(self, *k, **kw):
    buf = []
    for key in sorted(kw.keys()):
        v = kw[key]
        if isinstance(v, ConfigSet.ConfigSet):
            continue
        elif hasattr(v, '__call__'):
            buf.append(Utils.h_fun(v))
        else:
            buf.append(str(v))
    h = Utils.h_list(buf)
    dir = self.bldnode.abspath() + os.sep + (not Utils.is_win32 and '.' or '') + 'conf_check_' + Utils.to_hex(h)
    cachemode = kw.get('confcache', getattr(Options.options, 'confcache', None))
    if not cachemode and os.path.exists(dir):
        shutil.rmtree(dir)
    try:
        os.makedirs(dir)
    except OSError:
        pass
    try:
        os.stat(dir)
    except OSError:
        self.fatal('cannot use the configuration test folder %r' % dir)
    if cachemode == 1:
        try:
            proj = ConfigSet.ConfigSet(os.path.join(dir, 'cache_run_build'))
        except EnvironmentError:
            pass
        else:
            ret = proj['cache_run_build']
            if isinstance(ret, str) and ret.startswith('Test does not build'):
                self.fatal(ret)
            return ret
    bdir = os.path.join(dir, 'testbuild')
    if not os.path.exists(bdir):
        os.makedirs(bdir)
    cls_name = kw.get('run_build_cls') or getattr(self, 'run_build_cls', 'build')
    self.test_bld = bld = Context.create_context(cls_name, top_dir=dir, out_dir=bdir)
    bld.init_dirs()
    bld.progress_bar = 0
    bld.targets = '*'
    bld.logger = self.logger
    bld.all_envs.update(self.all_envs)
    bld.env = kw['env']
    bld.kw = kw
    bld.conf = self
    kw['build_fun'](bld)
    ret = -1
    try:
        try:
            bld.compile()
        except Errors.WafError:
            ret = 'Test does not build: %s' % traceback.format_exc()
            self.fatal(ret)
        else:
            ret = getattr(bld, 'retval', 0)
    finally:
        if cachemode:
            proj = ConfigSet.ConfigSet()
            proj['cache_run_build'] = ret
            proj.store(os.path.join(dir, 'cache_run_build'))
        else:
            shutil.rmtree(dir)
    return ret


@conf
def ret_msg(self, msg, args):
    if isinstance(msg, str):
        return msg
    return msg(args)


@conf
def test(self, *k, **kw):
    if not 'env' in kw:
        kw['env'] = self.env.derive()
    if kw.get('validate'):
        kw['validate'](kw)
    self.start_msg(kw['msg'], **kw)
    ret = None
    try:
        ret = self.run_build(*k, **kw)
    except self.errors.ConfigurationError:
        self.end_msg(kw['errmsg'], 'YELLOW', **kw)
        if Logs.verbose > 1:
            raise
        else:
            self.fatal('The configuration failed')
    else:
        kw['success'] = ret
    if kw.get('post_check'):
        ret = kw['post_check'](kw)
    if ret:
        self.end_msg(kw['errmsg'], 'YELLOW', **kw)
        self.fatal('The configuration failed %r' % ret)
    else:
        self.end_msg(self.ret_msg(kw['okmsg'], kw), **kw)
    return ret



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import itertools
import textwrap
import types

import pytest

from PyInstaller import HOMEPATH
from PyInstaller.depend import analysis
from PyInstaller.lib.modulegraph import modulegraph
import PyInstaller.log as logging
from PyInstaller.utils.tests import gen_sourcefile


def test_get_co_using_ctypes(tmp_path):
    logging.logger.setLevel(logging.DEBUG)
    mg = analysis.PyiModuleGraph(HOMEPATH, excludes=["xencodings"])
    script = tmp_path / 'script.py'
    script.write_text("import ctypes", encoding='utf-8')
    script_filename = str(script)
    mg.add_script(script_filename)
    res = mg.get_code_using("ctypes")
    # Script's code object must be in the results
    assert script_filename in res
    assert isinstance(res[script_filename], types.CodeType), res


def test_get_co_using_ctypes_from_extension():
    # If an extension module has an hidden import to ctypes (e.g. added by the hook), the extension module must not
    # show up in the result of `get_co_using_ctypes()`, since it has no code-object to be analyzed.
    # See issue #2492 and test_regression::issue_2492.
    logging.logger.setLevel(logging.DEBUG)
    mg = analysis.PyiModuleGraph(HOMEPATH, excludes=["xencodings"])
    struct = mg.createNode(modulegraph.Extension, '_struct', 'struct.so')
    mg.implyNodeReference(struct, 'ctypes')  # simulate the hidden import
    res = mg.get_code_using("ctypes")
    # _struct must not be in the results
    assert '_struct' not in res


def test_metadata_collection(tmp_path):
    from PyInstaller.utils.hooks import copy_metadata

    mg = analysis.PyiModuleGraph(HOMEPATH, excludes=["xencodings"])

    source = textwrap.dedent(
        '''
        from importlib.metadata import distribution, version
        import importlib.metadata

        distribution("setuptools")
        importlib.metadata.version("altgraph")
        '''
    )
    script = tmp_path / 'script.py'
    script.write_text(source, encoding='utf-8')

    mg.add_script(str(script))
    metadata = mg.metadata_required()

    assert copy_metadata("setuptools")[0] in metadata
    assert copy_metadata("altgraph")[0] in metadata


class FakePyiModuleGraph(analysis.PyiModuleGraph):
    def _analyze_base_modules(self):
        # suppress this to speed up set-up
        self._base_modules = ()


@pytest.fixture
def fresh_pyi_modgraph(monkeypatch):
    """
    Get a fresh PyiModuleGraph
    """
    def fake_base_modules(self):
        # speed up set up
        self._base_modules = ()

    logging.logger.setLevel(logging.DEBUG)
    # ensure we get a fresh PyiModuleGraph
    monkeypatch.setattr(analysis, "_cached_module_graph_", None)
    # speed up setup
    monkeypatch.setattr(analysis.PyiModuleGraph, "_analyze_base_modules", fake_base_modules)
    return analysis.initialize_modgraph()


def test_cached_graph_is_not_leaking(fresh_pyi_modgraph, monkeypatch, tmp_path):
    """
    Ensure cached PyiModulegraph can separate imports between scripts.
    """
    mg = fresh_pyi_modgraph
    # self-test 1: uuid is not included in the graph by default
    src = gen_sourcefile(tmp_path, """print""", test_id="1")
    mg.add_script(str(src))
    assert not mg.find_node("uuid")  # self-test

    # self-test 2: uuid is available and included when imported
    src = gen_sourcefile(tmp_path, """import uuid""", test_id="2")
    node = mg.add_script(str(src))
    assert node is not None
    names = [n.identifier for n in mg.iter_graph(start=node)]
    assert "uuid" in names

    # the actual test: uuid is not leaking to the other script
    src = gen_sourcefile(tmp_path, """print""", test_id="3")
    node = mg.add_script(str(src))
    assert node is not None
    names = [n.identifier for n in mg.iter_graph(start=node)]
    assert "uuid" not in names


def test_cached_graph_is_not_leaking_hidden_imports(fresh_pyi_modgraph, tmp_path):
    """
    Ensure cached PyiModulegraph can separate hidden imports between scripts.
    """
    mg = fresh_pyi_modgraph
    # self-test 1: skipped here, see test_cached_graph_is_not_leaking

    # self-test 2: uuid is included when hidden imported
    src = gen_sourcefile(tmp_path, """print""", test_id="2")
    node = mg.add_script(str(src))
    assert node is not None
    mg.add_hiddenimports(["uuid"])
    names = [n.identifier for n in mg.iter_graph(start=node)]
    assert "uuid" in names

    # the actual test: uuid is not leaking to the other script
    src = gen_sourcefile(tmp_path, """print""", test_id="3")
    node = mg.add_script(str(src))
    assert node is not None
    names = [n.identifier for n in mg.iter_graph(start=node)]
    assert "uuid" not in names


def test_graph_collects_script_dependencies(fresh_pyi_modgraph, tmp_path):
    mg = fresh_pyi_modgraph
    # self-test 1: uuid is not included in the graph by default
    src1 = gen_sourcefile(tmp_path, """print""", test_id="1")
    node = mg.add_script(str(src1))
    assert node is not None
    assert not mg.find_node("uuid")  # self-test

    # Add script importing uuid
    src2 = gen_sourcefile(tmp_path, """import uuid""", test_id="2")
    mg.add_script(str(src2))
    assert mg.find_node("uuid")  # self-test

    # The actual test: uuid is (indirectly) linked to the first script
    names = [n.identifier for n in mg.iter_graph(start=node)]
    assert str(src2) in names
    assert "uuid" in names


def _gen_pseudo_rthooks(name, rthook_dat, tmp_path, gen_files=True):
    # Create hooks directory
    hooks_dir = tmp_path / name
    rthooks_dir = hooks_dir / 'rthooks'
    rthooks_dir.mkdir(parents=True)

    # Create hook files
    if gen_files:
        for hook_file in itertools.chain(*rthook_dat.values()):
            (rthooks_dir / hook_file).touch()  # Create empty hook file

    # Create rthooks.dat file
    (hooks_dir / 'rthooks.dat').write_text(repr(rthook_dat), encoding='utf-8')
    return hooks_dir


def test_collect_rthooks_1(tmp_path, monkeypatch):
    rh1 = {"test_pyimodulegraph_mymod1": ["m1.py"]}
    hd1 = _gen_pseudo_rthooks("h1", rh1, tmp_path)
    mg = FakePyiModuleGraph(
        HOMEPATH,
        user_hook_dirs=[
            (str(hd1), analysis.HOOK_PRIORITY_BUILTIN_HOOKS),
        ],
    )
    assert len(mg._available_rthooks["test_pyimodulegraph_mymod1"]) == 1


def test_collect_rthooks_2(tmp_path, monkeypatch):
    rh1 = {"test_pyimodulegraph_mymod1": ["m1.py"]}
    rh2 = {"test_pyimodulegraph_mymod2": ["rth1.py", "rth1.py"]}
    hd1 = _gen_pseudo_rthooks("h1", rh1, tmp_path)
    hd2 = _gen_pseudo_rthooks("h2", rh2, tmp_path)
    mg = FakePyiModuleGraph(
        HOMEPATH,
        user_hook_dirs=[
            (str(hd1), analysis.HOOK_PRIORITY_BUILTIN_HOOKS),
            (str(hd2), analysis.HOOK_PRIORITY_BUILTIN_HOOKS),
        ],
    )
    assert len(mg._available_rthooks["test_pyimodulegraph_mymod1"]) == 1
    assert len(mg._available_rthooks["test_pyimodulegraph_mymod2"]) == 2


def test_collect_rthooks_3(tmp_path, monkeypatch):
    rh1 = {"test_pyimodulegraph_mymod1": ["m1.py"]}
    rh2 = {"test_pyimodulegraph_mymod1": ["rth1.py", "rth1.py"]}
    hd1 = _gen_pseudo_rthooks("h1", rh1, tmp_path)
    hd2 = _gen_pseudo_rthooks("h2", rh2, tmp_path)
    mg = FakePyiModuleGraph(
        HOMEPATH,
        user_hook_dirs=[
            (str(hd1), analysis.HOOK_PRIORITY_BUILTIN_HOOKS),
            (str(hd2), analysis.HOOK_PRIORITY_BUILTIN_HOOKS),
        ],
    )
    assert len(mg._available_rthooks["test_pyimodulegraph_mymod1"]) == 1


def test_collect_rthooks_fail_1(tmp_path, monkeypatch):
    rh1 = {"test_pyimodulegraph_mymod1": ["m1.py"]}
    hd1 = _gen_pseudo_rthooks("h1", rh1, tmp_path, False)
    with pytest.raises(AssertionError):
        FakePyiModuleGraph(
            HOMEPATH,
            user_hook_dirs=[
                (str(hd1), analysis.HOOK_PRIORITY_BUILTIN_HOOKS),
            ],
        )


class FakeGraph(analysis.PyiModuleGraph):
    """
    A simplified module graph containing a single node module *foo* with user-defined content.
    """
    def __init__(self, source):
        self.code = compile(source, "<>", "exec")

    def get_code_using(self, package):
        return {"foo": self.code}


def test_metadata_searching():
    """
    Test the top level for bytecode scanning for metadata requirements.
    """
    from PyInstaller.utils.hooks import copy_metadata

    # This test analyses code which implies that PyInstaller's own metadata (and possibly that of its dependencies) is
    # required.
    pyinstaller = set(copy_metadata("pyinstaller"))
    with_dependencies = set(copy_metadata("pyinstaller", recursive=True))

    self = FakeGraph("from importlib.metadata import distribution; distribution('pyinstaller')")
    assert pyinstaller == self.metadata_required()

    self = FakeGraph("import pkg_resources; pkg_resources.get_distribution('pyinstaller')")
    assert pyinstaller == self.metadata_required()

    self = FakeGraph("import pkg_resources; pkg_resources.require('pyinstaller')")
    assert with_dependencies == self.metadata_required()



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib.Tools import ccroot, ar, gxx
from waflib.Configure import conf


@conf
def find_clangxx(conf):
    cxx = conf.find_program('clang++', var='CXX')
    conf.get_cc_version(cxx, clang=True)
    conf.env.CXX_NAME = 'clang'


def configure(conf):
    conf.find_clangxx()
    conf.find_program(['llvm-ar', 'ar'], var='AR')
    conf.find_ar()
    conf.gxx_common_flags()
    conf.gxx_modifier_platform()
    conf.cxx_load_tools()
    conf.cxx_add_flags()
    conf.link_add_flags()



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re
from waflib import Build, Context, Errors, Logs, Node, Options, Task, Utils
from waflib.TaskGen import extension, taskgen_method
from waflib.Configure import conf


class valac(Task.Task):
    vars = ["VALAC", "VALAC_VERSION", "VALAFLAGS"]
    ext_out = ['.h']

    def run(self):
        cmd = self.env.VALAC + self.env.VALAFLAGS
        resources = getattr(self, 'vala_exclude', [])
        cmd.extend([a.abspath() for a in self.inputs if a not in resources])
        ret = self.exec_command(cmd, cwd=self.vala_dir_node.abspath())
        if ret:
            return ret
        if self.generator.dump_deps_node:
            self.generator.dump_deps_node.write('\n'.join(self.generator.packages))
        return ret


@taskgen_method
def init_vala_task(self):
    self.profile = getattr(self, 'profile', 'gobject')
    self.packages = packages = Utils.to_list(getattr(self, 'packages', []))
    self.use = Utils.to_list(getattr(self, 'use', []))
    if packages and not self.use:
        self.use = packages[:]
    if self.profile == 'gobject':
        if not 'GOBJECT' in self.use:
            self.use.append('GOBJECT')

    def addflags(flags):
        self.env.append_value('VALAFLAGS', flags)

    if self.profile:
        addflags('--profile=%s' % self.profile)
    valatask = self.valatask
    if hasattr(self, 'vala_dir'):
        if isinstance(self.vala_dir, str):
            valatask.vala_dir_node = self.path.get_bld().make_node(self.vala_dir)
            try:
                valatask.vala_dir_node.mkdir()
            except OSError:
                raise self.bld.fatal('Cannot create the vala dir %r' % valatask.vala_dir_node)
        else:
            valatask.vala_dir_node = self.vala_dir
    else:
        valatask.vala_dir_node = self.path.get_bld()
    addflags('--directory=%s' % valatask.vala_dir_node.abspath())
    if hasattr(self, 'thread'):
        if self.profile == 'gobject':
            if not 'GTHREAD' in self.use:
                self.use.append('GTHREAD')
        else:
            Logs.warn('Profile %s means no threading support', self.profile)
            self.thread = False
        if self.thread:
            addflags('--thread')
    self.is_lib = 'cprogram' not in self.features
    if self.is_lib:
        addflags('--library=%s' % self.target)
        h_node = valatask.vala_dir_node.find_or_declare('%s.h' % self.target)
        valatask.outputs.append(h_node)
        addflags('--header=%s' % h_node.name)
        valatask.outputs.append(valatask.vala_dir_node.find_or_declare('%s.vapi' % self.target))
        if getattr(self, 'gir', None):
            gir_node = valatask.vala_dir_node.find_or_declare('%s.gir' % self.gir)
            addflags('--gir=%s' % gir_node.name)
            valatask.outputs.append(gir_node)
    self.vala_target_glib = getattr(self, 'vala_target_glib', getattr(Options.options, 'vala_target_glib', None))
    if self.vala_target_glib:
        addflags('--target-glib=%s' % self.vala_target_glib)
    addflags(['--define=%s' % x for x in Utils.to_list(getattr(self, 'vala_defines', []))])
    packages_private = Utils.to_list(getattr(self, 'packages_private', []))
    addflags(['--pkg=%s' % x for x in packages_private])

    def _get_api_version():
        api_version = '1.0'
        if hasattr(Context.g_module, 'API_VERSION'):
            version = Context.g_module.API_VERSION.split(".")
            if version[0] == "0":
                api_version = "0." + version[1]
            else:
                api_version = version[0] + ".0"
        return api_version

    self.includes = Utils.to_list(getattr(self, 'includes', []))
    valatask.install_path = getattr(self, 'install_path', '')
    valatask.vapi_path = getattr(self, 'vapi_path', '${DATAROOTDIR}/vala/vapi')
    valatask.pkg_name = getattr(self, 'pkg_name', self.env.PACKAGE)
    valatask.header_path = getattr(self, 'header_path', '${INCLUDEDIR}/%s-%s' % (valatask.pkg_name, _get_api_version()))
    valatask.install_binding = getattr(self, 'install_binding', True)
    self.vapi_dirs = vapi_dirs = Utils.to_list(getattr(self, 'vapi_dirs', []))
    if hasattr(self, 'use'):
        local_packages = Utils.to_list(self.use)[:]
        seen = []
        while len(local_packages) > 0:
            package = local_packages.pop()
            if package in seen:
                continue
            seen.append(package)
            try:
                package_obj = self.bld.get_tgen_by_name(package)
            except Errors.WafError:
                continue
            package_obj.post()
            package_name = package_obj.target
            task = getattr(package_obj, 'valatask', None)
            if task:
                for output in task.outputs:
                    if output.name == package_name + ".vapi":
                        valatask.set_run_after(task)
                        if package_name not in packages:
                            packages.append(package_name)
                        if output.parent not in vapi_dirs:
                            vapi_dirs.append(output.parent)
                        if output.parent not in self.includes:
                            self.includes.append(output.parent)
            if hasattr(package_obj, 'use'):
                lst = self.to_list(package_obj.use)
                lst.reverse()
                local_packages = [pkg for pkg in lst if pkg not in seen] + local_packages
    addflags(['--pkg=%s' % p for p in packages])
    for vapi_dir in vapi_dirs:
        if isinstance(vapi_dir, Node.Node):
            v_node = vapi_dir
        else:
            v_node = self.path.find_dir(vapi_dir)
        if not v_node:
            Logs.warn('Unable to locate Vala API directory: %r', vapi_dir)
        else:
            addflags('--vapidir=%s' % v_node.abspath())
    self.dump_deps_node = None
    if self.is_lib and self.packages:
        self.dump_deps_node = valatask.vala_dir_node.find_or_declare('%s.deps' % self.target)
        valatask.outputs.append(self.dump_deps_node)
    if self.is_lib and valatask.install_binding:
        headers_list = [o for o in valatask.outputs if o.suffix() == ".h"]
        if headers_list:
            self.install_vheader = self.add_install_files(install_to=valatask.header_path, install_from=headers_list)
        vapi_list = [o for o in valatask.outputs if (o.suffix() in (".vapi", ".deps"))]
        if vapi_list:
            self.install_vapi = self.add_install_files(install_to=valatask.vapi_path, install_from=vapi_list)
        gir_list = [o for o in valatask.outputs if o.suffix() == '.gir']
        if gir_list:
            self.install_gir = self.add_install_files(
                install_to=getattr(self, 'gir_path', '${DATAROOTDIR}/gir-1.0'), install_from=gir_list
            )
    if hasattr(self, 'vala_resources'):
        nodes = self.to_nodes(self.vala_resources)
        valatask.vala_exclude = getattr(valatask, 'vala_exclude', []) + nodes
        valatask.inputs.extend(nodes)
        for x in nodes:
            addflags(['--gresources', x.abspath()])


@extension('.vala', '.gs')
def vala_file(self, node):
    try:
        valatask = self.valatask
    except AttributeError:
        valatask = self.valatask = self.create_task('valac')
        self.init_vala_task()
    valatask.inputs.append(node)
    name = node.name[:node.name.rfind('.')] + '.c'
    c_node = valatask.vala_dir_node.find_or_declare(name)
    valatask.outputs.append(c_node)
    self.source.append(c_node)


@extension('.vapi')
def vapi_file(self, node):
    try:
        valatask = self.valatask
    except AttributeError:
        valatask = self.valatask = self.create_task('valac')
        self.init_vala_task()
    valatask.inputs.append(node)


@conf
def find_valac(self, valac_name, min_version):
    valac = self.find_program(valac_name, var='VALAC')
    try:
        output = self.cmd_and_log(valac + ['--version'])
    except Errors.WafError:
        valac_version = None
    else:
        ver = re.search(r'\d+.\d+.\d+', output).group().split('.')
        valac_version = tuple([int(x) for x in ver])
    self.msg(
        'Checking for %s version >= %r' % (valac_name, min_version), valac_version, valac_version
        and valac_version >= min_version
    )
    if valac and valac_version < min_version:
        self.fatal("%s version %r is too old, need >= %r" % (valac_name, valac_version, min_version))
    self.env.VALAC_VERSION = valac_version
    return valac


@conf
def check_vala(self, min_version=(0, 8, 0), branch=None):
    if self.env.VALA_MINVER:
        min_version = self.env.VALA_MINVER
    if self.env.VALA_MINVER_BRANCH:
        branch = self.env.VALA_MINVER_BRANCH
    if not branch:
        branch = min_version[:2]
    try:
        find_valac(self, 'valac-%d.%d' % (branch[0], branch[1]), min_version)
    except self.errors.ConfigurationError:
        find_valac(self, 'valac', min_version)


@conf
def check_vala_deps(self):
    if not self.env.HAVE_GOBJECT:
        pkg_args = {'package': 'gobject-2.0', 'uselib_store': 'GOBJECT', 'args': '--cflags --libs'}
        if getattr(Options.options, 'vala_target_glib', None):
            pkg_args['atleast_version'] = Options.options.vala_target_glib
        self.check_cfg(**pkg_args)
    if not self.env.HAVE_GTHREAD:
        pkg_args = {'package': 'gthread-2.0', 'uselib_store': 'GTHREAD', 'args': '--cflags --libs'}
        if getattr(Options.options, 'vala_target_glib', None):
            pkg_args['atleast_version'] = Options.options.vala_target_glib
        self.check_cfg(**pkg_args)


def configure(self):
    self.load('gnu_dirs')
    self.check_vala_deps()
    self.check_vala()
    self.add_os_flags('VALAFLAGS')
    self.env.append_unique('VALAFLAGS', ['-C'])


def options(opt):
    opt.load('gnu_dirs')
    valaopts = opt.add_option_group('Vala Compiler Options')
    valaopts.add_option(
        '--vala-target-glib',
        default=None,
        dest='vala_target_glib',
        metavar='MAJOR.MINOR',
        help='Target version of glib for Vala GObject code generation'
    )



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re, sys, tempfile, traceback
from waflib import Utils, Logs, Errors

NOT_RUN = 0
MISSING = 1
CRASHED = 2
EXCEPTION = 3
CANCELED = 4
SKIPPED = 8
SUCCESS = 9
ASK_LATER = -1
SKIP_ME = -2
RUN_ME = -3
CANCEL_ME = -4
COMPILE_TEMPLATE_SHELL = '''
def f(tsk):
	env = tsk.env
	gen = tsk.generator
	bld = gen.bld
	cwdx = tsk.get_cwd()
	p = env.get_flat
	def to_list(xx):
		if isinstance(xx, str): return [xx]
		return xx
	tsk.last_cmd = cmd = \'\'\' %s \'\'\' % s
	return tsk.exec_command(cmd, cwd=cwdx, env=env.env or None)
'''
COMPILE_TEMPLATE_NOSHELL = '''
def f(tsk):
	env = tsk.env
	gen = tsk.generator
	bld = gen.bld
	cwdx = tsk.get_cwd()
	def to_list(xx):
		if isinstance(xx, str): return [xx]
		return xx
	def merge(lst1, lst2):
		if lst1 and lst2:
			return lst1[:-1] + [lst1[-1] + lst2[0]] + lst2[1:]
		return lst1 + lst2
	lst = []
	%s
	if '' in lst:
		lst = [x for x in lst if x]
	tsk.last_cmd = lst
	return tsk.exec_command(lst, cwd=cwdx, env=env.env or None)
'''
COMPILE_TEMPLATE_SIG_VARS = '''
def f(tsk):
	sig = tsk.generator.bld.hash_env_vars(tsk.env, tsk.vars)
	tsk.m.update(sig)
	env = tsk.env
	gen = tsk.generator
	bld = gen.bld
	cwdx = tsk.get_cwd()
	p = env.get_flat
	buf = []
	%s
	tsk.m.update(repr(buf).encode())
'''
classes = {}


class store_task_type(type):
    def __init__(cls, name, bases, dict):
        super(store_task_type, cls).__init__(name, bases, dict)
        name = cls.__name__
        if name != 'evil' and name != 'Task':
            if getattr(cls, 'run_str', None):
                (f, dvars) = compile_fun(cls.run_str, cls.shell)
                cls.hcode = Utils.h_cmd(cls.run_str)
                cls.orig_run_str = cls.run_str
                cls.run_str = None
                cls.run = f
                cls.vars = list(set(cls.vars + dvars))
                cls.vars.sort()
                if cls.vars:
                    fun = compile_sig_vars(cls.vars)
                    if fun:
                        cls.sig_vars = fun
            elif getattr(cls, 'run', None) and not 'hcode' in cls.__dict__:
                cls.hcode = Utils.h_cmd(cls.run)
            getattr(cls, 'register', classes)[name] = cls


evil = store_task_type('evil', (object,), {})


class Task(evil):
    vars = []
    always_run = False
    shell = False
    color = 'GREEN'
    ext_in = []
    ext_out = []
    before = []
    after = []
    hcode = Utils.SIG_NIL
    keep_last_cmd = False
    weight = 0
    tree_weight = 0
    prio_order = 0
    __slots__ = ('hasrun', 'generator', 'env', 'inputs', 'outputs', 'dep_nodes', 'run_after')

    def __init__(self, *k, **kw):
        self.hasrun = NOT_RUN
        try:
            self.generator = kw['generator']
        except KeyError:
            self.generator = self
        self.env = kw['env']
        self.inputs = []
        self.outputs = []
        self.dep_nodes = []
        self.run_after = set()

    def __lt__(self, other):
        return self.priority() > other.priority()

    def __le__(self, other):
        return self.priority() >= other.priority()

    def __gt__(self, other):
        return self.priority() < other.priority()

    def __ge__(self, other):
        return self.priority() <= other.priority()

    def get_cwd(self):
        bld = self.generator.bld
        ret = getattr(self, 'cwd', None) or getattr(bld, 'cwd', bld.bldnode)
        if isinstance(ret, str):
            if os.path.isabs(ret):
                ret = bld.root.make_node(ret)
            else:
                ret = self.generator.path.make_node(ret)
        return ret

    def quote_flag(self, x):
        old = x
        if '\\' in x:
            x = x.replace('\\', '\\\\')
        if '"' in x:
            x = x.replace('"', '\\"')
        if old != x or ' ' in x or '\t' in x or "'" in x:
            x = '"%s"' % x
        return x

    def priority(self):
        return (self.weight + self.prio_order, -getattr(self.generator, 'tg_idx_count', 0))

    def split_argfile(self, cmd):
        return ([cmd[0]], [self.quote_flag(x) for x in cmd[1:]])

    def exec_command(self, cmd, **kw):
        if not 'cwd' in kw:
            kw['cwd'] = self.get_cwd()
        if hasattr(self, 'timeout'):
            kw['timeout'] = self.timeout
        if self.env.PATH:
            env = kw['env'] = dict(kw.get('env') or self.env.env or os.environ)
            env['PATH'] = self.env.PATH if isinstance(self.env.PATH, str) else os.pathsep.join(self.env.PATH)
        if hasattr(self, 'stdout'):
            kw['stdout'] = self.stdout
        if hasattr(self, 'stderr'):
            kw['stderr'] = self.stderr
        if not isinstance(cmd, str):
            if Utils.is_win32:
                too_long = sum([len(arg) for arg in cmd]) + len(cmd) > 8192
            else:
                too_long = len(cmd) > 200000
            if too_long and getattr(self, 'allow_argsfile', True):
                cmd, args = self.split_argfile(cmd)
                try:
                    (fd, tmp) = tempfile.mkstemp()
                    os.write(fd, '\r\n'.join(args).encode())
                    os.close(fd)
                    if Logs.verbose:
                        Logs.debug('argfile: @%r -> %r', tmp, args)
                    return self.generator.bld.exec_command(cmd + ['@' + tmp], **kw)
                finally:
                    try:
                        os.remove(tmp)
                    except OSError:
                        pass
        return self.generator.bld.exec_command(cmd, **kw)

    def process(self):
        try:
            del self.generator.bld.task_sigs[self.uid()]
        except KeyError:
            pass
        try:
            ret = self.run()
        except Exception:
            self.err_msg = traceback.format_exc()
            self.hasrun = EXCEPTION
        else:
            if ret:
                self.err_code = ret
                self.hasrun = CRASHED
            else:
                try:
                    self.post_run()
                except Errors.WafError:
                    pass
                except Exception:
                    self.err_msg = traceback.format_exc()
                    self.hasrun = EXCEPTION
                else:
                    self.hasrun = SUCCESS
        if self.hasrun != SUCCESS and self.scan:
            try:
                del self.generator.bld.imp_sigs[self.uid()]
            except KeyError:
                pass

    def log_display(self, bld):
        if self.generator.bld.progress_bar == 3:
            return
        s = self.display()
        if s:
            if bld.logger:
                logger = bld.logger
            else:
                logger = Logs
            if self.generator.bld.progress_bar == 1:
                c1 = Logs.colors.cursor_off
                c2 = Logs.colors.cursor_on
                logger.info(s, extra={'stream': sys.stderr, 'terminator': '', 'c1': c1, 'c2': c2})
            else:
                logger.info(s, extra={'terminator': '', 'c1': '', 'c2': ''})

    def display(self):
        col1 = Logs.colors(self.color)
        col2 = Logs.colors.NORMAL
        master = self.generator.bld.producer

        def cur():
            return master.processed - master.ready.qsize()

        if self.generator.bld.progress_bar == 1:
            return self.generator.bld.progress_line(cur(), master.total, col1, col2)
        if self.generator.bld.progress_bar == 2:
            ela = str(self.generator.bld.timer)
            try:
                ins = ','.join([n.name for n in self.inputs])
            except AttributeError:
                ins = ''
            try:
                outs = ','.join([n.name for n in self.outputs])
            except AttributeError:
                outs = ''
            return '|Total %s|Current %s|Inputs %s|Outputs %s|Time %s|\n' % (master.total, cur(), ins, outs, ela)
        s = str(self)
        if not s:
            return None
        total = master.total
        n = len(str(total))
        fs = '[%%%dd/%%%dd] %%s%%s%%s%%s\n' % (n, n)
        kw = self.keyword()
        if kw:
            kw += ' '
        return fs % (cur(), total, kw, col1, s, col2)

    def hash_constraints(self):
        return (
            tuple(self.before), tuple(self.after), tuple(self.ext_in), tuple(self.ext_out), self.__class__.__name__,
            self.hcode
        )

    def format_error(self):
        if Logs.verbose:
            msg = ': %r\n%r' % (self, getattr(self, 'last_cmd', ''))
        else:
            msg = ' (run with -v to display more information)'
        name = getattr(self.generator, 'name', '')
        if getattr(self, "err_msg", None):
            return self.err_msg
        elif not self.hasrun:
            return 'task in %r was not executed for some reason: %r' % (name, self)
        elif self.hasrun == CRASHED:
            try:
                return ' -> task in %r failed with exit status %r%s' % (name, self.err_code, msg)
            except AttributeError:
                return ' -> task in %r failed%s' % (name, msg)
        elif self.hasrun == MISSING:
            return ' -> missing files in %r%s' % (name, msg)
        elif self.hasrun == CANCELED:
            return ' -> %r canceled because of missing dependencies' % name
        else:
            return 'invalid status for task in %r: %r' % (name, self.hasrun)

    def colon(self, var1, var2):
        tmp = self.env[var1]
        if not tmp:
            return []
        if isinstance(var2, str):
            it = self.env[var2]
        else:
            it = var2
        if isinstance(tmp, str):
            return [tmp % x for x in it]
        else:
            lst = []
            for y in it:
                lst.extend(tmp)
                lst.append(y)
            return lst

    def __str__(self):
        name = self.__class__.__name__
        if self.outputs:
            if name.endswith(('lib', 'program')) or not self.inputs:
                node = self.outputs[0]
                return node.path_from(node.ctx.launch_node())
        if not (self.inputs or self.outputs):
            return self.__class__.__name__
        if len(self.inputs) == 1:
            node = self.inputs[0]
            return node.path_from(node.ctx.launch_node())
        src_str = ' '.join([a.path_from(a.ctx.launch_node()) for a in self.inputs])
        tgt_str = ' '.join([a.path_from(a.ctx.launch_node()) for a in self.outputs])
        if self.outputs:
            sep = ' -> '
        else:
            sep = ''
        return '%s: %s%s%s' % (self.__class__.__name__, src_str, sep, tgt_str)

    def keyword(self):
        name = self.__class__.__name__
        if name.endswith(('lib', 'program')):
            return 'Linking'
        if len(self.inputs) == 1 and len(self.outputs) == 1:
            return 'Compiling'
        if not self.inputs:
            if self.outputs:
                return 'Creating'
            else:
                return 'Running'
        return 'Processing'

    def __repr__(self):
        try:
            ins = ",".join([x.name for x in self.inputs])
            outs = ",".join([x.name for x in self.outputs])
        except AttributeError:
            ins = ",".join([str(x) for x in self.inputs])
            outs = ",".join([str(x) for x in self.outputs])
        return "".join(['\n\t{task %r: ' % id(self), self.__class__.__name__, " ", ins, " -> ", outs, '}'])

    def uid(self):
        try:
            return self.uid_
        except AttributeError:
            m = Utils.md5(self.__class__.__name__)
            up = m.update
            for x in self.inputs + self.outputs:
                up(x.abspath())
            self.uid_ = m.digest()
            return self.uid_

    def set_inputs(self, inp):
        if isinstance(inp, list):
            self.inputs += inp
        else:
            self.inputs.append(inp)

    def set_outputs(self, out):
        if isinstance(out, list):
            self.outputs += out
        else:
            self.outputs.append(out)

    def set_run_after(self, task):
        assert isinstance(task, Task)
        self.run_after.add(task)

    def signature(self):
        try:
            return self.cache_sig
        except AttributeError:
            pass
        self.m = Utils.md5(self.hcode)
        self.sig_explicit_deps()
        self.sig_vars()
        if self.scan:
            try:
                self.sig_implicit_deps()
            except Errors.TaskRescan:
                return self.signature()
        ret = self.cache_sig = self.m.digest()
        return ret

    def runnable_status(self):
        bld = self.generator.bld
        if bld.is_install < 0:
            return SKIP_ME
        for t in self.run_after:
            if not t.hasrun:
                return ASK_LATER
            elif t.hasrun < SKIPPED:
                return CANCEL_ME
        try:
            new_sig = self.signature()
        except Errors.TaskNotReady:
            return ASK_LATER
        key = self.uid()
        try:
            prev_sig = bld.task_sigs[key]
        except KeyError:
            Logs.debug('task: task %r must run: it was never run before or the task code changed', self)
            return RUN_ME
        if new_sig != prev_sig:
            Logs.debug('task: task %r must run: the task signature changed', self)
            return RUN_ME
        for node in self.outputs:
            sig = bld.node_sigs.get(node)
            if not sig:
                Logs.debug('task: task %r must run: an output node has no signature', self)
                return RUN_ME
            if sig != key:
                Logs.debug('task: task %r must run: an output node was produced by another task', self)
                return RUN_ME
            if not node.exists():
                Logs.debug('task: task %r must run: an output node does not exist', self)
                return RUN_ME
        return (self.always_run and RUN_ME) or SKIP_ME

    def post_run(self):
        bld = self.generator.bld
        for node in self.outputs:
            if not node.exists():
                self.hasrun = MISSING
                self.err_msg = '-> missing file: %r' % node.abspath()
                raise Errors.WafError(self.err_msg)
            bld.node_sigs[node] = self.uid()
        bld.task_sigs[self.uid()] = self.signature()
        if not self.keep_last_cmd:
            try:
                del self.last_cmd
            except AttributeError:
                pass

    def sig_explicit_deps(self):
        bld = self.generator.bld
        upd = self.m.update
        for x in self.inputs + self.dep_nodes:
            upd(x.get_bld_sig())
        if bld.deps_man:
            additional_deps = bld.deps_man
            for x in self.inputs + self.outputs:
                try:
                    d = additional_deps[x]
                except KeyError:
                    continue
                for v in d:
                    try:
                        v = v.get_bld_sig()
                    except AttributeError:
                        if hasattr(v, '__call__'):
                            v = v()
                    upd(v)

    def sig_deep_inputs(self):
        bld = self.generator.bld
        lst = [bld.task_sigs[bld.node_sigs[node]] for node in (self.inputs + self.dep_nodes) if node.is_bld()]
        self.m.update(Utils.h_list(lst))

    def sig_vars(self):
        sig = self.generator.bld.hash_env_vars(self.env, self.vars)
        self.m.update(sig)

    scan = None

    def sig_implicit_deps(self):
        bld = self.generator.bld
        key = self.uid()
        prev = bld.imp_sigs.get(key, [])
        if prev:
            try:
                if prev == self.compute_sig_implicit_deps():
                    return prev
            except Errors.TaskNotReady:
                raise
            except EnvironmentError:
                for x in bld.node_deps.get(self.uid(), []):
                    if not x.is_bld() and not x.exists():
                        try:
                            del x.parent.children[x.name]
                        except KeyError:
                            pass
            del bld.imp_sigs[key]
            raise Errors.TaskRescan('rescan')
        (bld.node_deps[key], bld.raw_deps[key]) = self.scan()
        if Logs.verbose:
            Logs.debug('deps: scanner for %s: %r; unresolved: %r', self, bld.node_deps[key], bld.raw_deps[key])
        try:
            bld.imp_sigs[key] = self.compute_sig_implicit_deps()
        except EnvironmentError:
            for k in bld.node_deps.get(self.uid(), []):
                if not k.exists():
                    Logs.warn(
                        'Dependency %r for %r is missing: check the task declaration and the build order!', k, self
                    )
            raise

    def compute_sig_implicit_deps(self):
        upd = self.m.update
        self.are_implicit_nodes_ready()
        for k in self.generator.bld.node_deps.get(self.uid(), []):
            upd(k.get_bld_sig())
        return self.m.digest()

    def are_implicit_nodes_ready(self):
        bld = self.generator.bld
        try:
            cache = bld.dct_implicit_nodes
        except AttributeError:
            bld.dct_implicit_nodes = cache = {}
        try:
            dct = cache[bld.current_group]
        except KeyError:
            dct = cache[bld.current_group] = {}
            for tsk in bld.cur_tasks:
                for x in tsk.outputs:
                    dct[x] = tsk
        modified = False
        for x in bld.node_deps.get(self.uid(), []):
            if x in dct:
                self.run_after.add(dct[x])
                modified = True
        if modified:
            for tsk in self.run_after:
                if not tsk.hasrun:
                    raise Errors.TaskNotReady('not ready')


if sys.hexversion > 0x3000000:

    def uid(self):
        try:
            return self.uid_
        except AttributeError:
            m = Utils.md5(self.__class__.__name__.encode('latin-1', 'xmlcharrefreplace'))
            up = m.update
            for x in self.inputs + self.outputs:
                up(x.abspath().encode('latin-1', 'xmlcharrefreplace'))
            self.uid_ = m.digest()
            return self.uid_

    uid.__doc__ = Task.uid.__doc__
    Task.uid = uid


def is_before(t1, t2):
    to_list = Utils.to_list
    for k in to_list(t2.ext_in):
        if k in to_list(t1.ext_out):
            return 1
    if t1.__class__.__name__ in to_list(t2.after):
        return 1
    if t2.__class__.__name__ in to_list(t1.before):
        return 1
    return 0


def set_file_constraints(tasks):
    ins = Utils.defaultdict(set)
    outs = Utils.defaultdict(set)
    for x in tasks:
        for a in x.inputs:
            ins[a].add(x)
        for a in x.dep_nodes:
            ins[a].add(x)
        for a in x.outputs:
            outs[a].add(x)
    links = set(ins.keys()).intersection(outs.keys())
    for k in links:
        for a in ins[k]:
            a.run_after.update(outs[k])


class TaskGroup(object):
    def __init__(self, prev, next):
        self.prev = prev
        self.next = next
        self.done = False

    def get_hasrun(self):
        for k in self.prev:
            if not k.hasrun:
                return NOT_RUN
        return SUCCESS

    hasrun = property(get_hasrun, None)


def set_precedence_constraints(tasks):
    cstr_groups = Utils.defaultdict(list)
    for x in tasks:
        h = x.hash_constraints()
        cstr_groups[h].append(x)
    keys = list(cstr_groups.keys())
    maxi = len(keys)
    for i in range(maxi):
        t1 = cstr_groups[keys[i]][0]
        for j in range(i + 1, maxi):
            t2 = cstr_groups[keys[j]][0]
            if is_before(t1, t2):
                a = i
                b = j
            elif is_before(t2, t1):
                a = j
                b = i
            else:
                continue
            a = cstr_groups[keys[a]]
            b = cstr_groups[keys[b]]
            if len(a) < 2 or len(b) < 2:
                for x in b:
                    x.run_after.update(a)
            else:
                group = TaskGroup(set(a), set(b))
                for x in b:
                    x.run_after.add(group)


def funex(c):
    dc = {}
    exec(c, dc)
    return dc['f']


re_cond = re.compile(r'(?P<var>\w+)|(?P<or>\|)|(?P<and>&)')
re_novar = re.compile(r'^(SRC|TGT)\W+.*?$')
reg_act = re.compile(r'(?P<backslash>\\)|(?P<dollar>\$\$)|(?P<subst>\$\{(?P<var>\w+)(?P<code>.*?)\})', re.M)


def compile_fun_shell(line):
    extr = []

    def repl(match):
        g = match.group
        if g('dollar'):
            return "$"
        elif g('backslash'):
            return '\\\\'
        elif g('subst'):
            extr.append((g('var'), g('code')))
            return "%s"
        return None

    line = reg_act.sub(repl, line) or line
    dvars = []

    def add_dvar(x):
        if x not in dvars:
            dvars.append(x)

    def replc(m):
        if m.group('and'):
            return ' and '
        elif m.group('or'):
            return ' or '
        else:
            x = m.group('var')
            add_dvar(x)
            return 'env[%r]' % x

    parm = []
    app = parm.append
    for (var, meth) in extr:
        if var == 'SRC':
            if meth:
                app('tsk.inputs%s' % meth)
            else:
                app('" ".join([a.path_from(cwdx) for a in tsk.inputs])')
        elif var == 'TGT':
            if meth:
                app('tsk.outputs%s' % meth)
            else:
                app('" ".join([a.path_from(cwdx) for a in tsk.outputs])')
        elif meth:
            if meth.startswith(':'):
                add_dvar(var)
                m = meth[1:]
                if m == 'SRC':
                    m = '[a.path_from(cwdx) for a in tsk.inputs]'
                elif m == 'TGT':
                    m = '[a.path_from(cwdx) for a in tsk.outputs]'
                elif re_novar.match(m):
                    m = '[tsk.inputs%s]' % m[3:]
                elif re_novar.match(m):
                    m = '[tsk.outputs%s]' % m[3:]
                else:
                    add_dvar(m)
                    if m[:3] not in ('tsk', 'gen', 'bld'):
                        m = '%r' % m
                app('" ".join(tsk.colon(%r, %s))' % (var, m))
            elif meth.startswith('?'):
                expr = re_cond.sub(replc, meth[1:])
                app('p(%r) if (%s) else ""' % (var, expr))
            else:
                call = '%s%s' % (var, meth)
                add_dvar(call)
                app(call)
        else:
            add_dvar(var)
            app("p('%s')" % var)
    if parm:
        parm = "%% (%s) " % (',\n\t\t'.join(parm))
    else:
        parm = ''
    c = COMPILE_TEMPLATE_SHELL % (line, parm)
    Logs.debug('action: %s', c.strip().splitlines())
    return (funex(c), dvars)


reg_act_noshell = re.compile(
    r"(?P<space>\s+)|(?P<subst>\$\{(?P<var>\w+)(?P<code>.*?)\})|(?P<text>([^$ \t\n\r\f\v]|\$\$)+)", re.M
)


def compile_fun_noshell(line):
    buf = []
    dvars = []
    merge = False
    app = buf.append

    def add_dvar(x):
        if x not in dvars:
            dvars.append(x)

    def replc(m):
        if m.group('and'):
            return ' and '
        elif m.group('or'):
            return ' or '
        else:
            x = m.group('var')
            add_dvar(x)
            return 'env[%r]' % x

    for m in reg_act_noshell.finditer(line):
        if m.group('space'):
            merge = False
            continue
        elif m.group('text'):
            app('[%r]' % m.group('text').replace('$$', '$'))
        elif m.group('subst'):
            var = m.group('var')
            code = m.group('code')
            if var == 'SRC':
                if code:
                    app('[tsk.inputs%s]' % code)
                else:
                    app('[a.path_from(cwdx) for a in tsk.inputs]')
            elif var == 'TGT':
                if code:
                    app('[tsk.outputs%s]' % code)
                else:
                    app('[a.path_from(cwdx) for a in tsk.outputs]')
            elif code:
                if code.startswith(':'):
                    add_dvar(var)
                    m = code[1:]
                    if m == 'SRC':
                        m = '[a.path_from(cwdx) for a in tsk.inputs]'
                    elif m == 'TGT':
                        m = '[a.path_from(cwdx) for a in tsk.outputs]'
                    elif re_novar.match(m):
                        m = '[tsk.inputs%s]' % m[3:]
                    elif re_novar.match(m):
                        m = '[tsk.outputs%s]' % m[3:]
                    else:
                        add_dvar(m)
                        if m[:3] not in ('tsk', 'gen', 'bld'):
                            m = '%r' % m
                    app('tsk.colon(%r, %s)' % (var, m))
                elif code.startswith('?'):
                    expr = re_cond.sub(replc, code[1:])
                    app('to_list(env[%r] if (%s) else [])' % (var, expr))
                else:
                    call = '%s%s' % (var, code)
                    add_dvar(call)
                    app('to_list(%s)' % call)
            else:
                app('to_list(env[%r])' % var)
                add_dvar(var)
        if merge:
            tmp = 'merge(%s, %s)' % (buf[-2], buf[-1])
            del buf[-1]
            buf[-1] = tmp
        merge = True
    buf = ['lst.extend(%s)' % x for x in buf]
    fun = COMPILE_TEMPLATE_NOSHELL % "\n\t".join(buf)
    Logs.debug('action: %s', fun.strip().splitlines())
    return (funex(fun), dvars)


def compile_fun(line, shell=False):
    if isinstance(line, str):
        if line.find('<') > 0 or line.find('>') > 0 or line.find('&&') > 0:
            shell = True
    else:
        dvars_lst = []
        funs_lst = []
        for x in line:
            if isinstance(x, str):
                fun, dvars = compile_fun(x, shell)
                dvars_lst += dvars
                funs_lst.append(fun)
            else:
                funs_lst.append(x)

        def composed_fun(task):
            for x in funs_lst:
                ret = x(task)
                if ret:
                    return ret
            return None

        return composed_fun, dvars_lst
    if shell:
        return compile_fun_shell(line)
    else:
        return compile_fun_noshell(line)


def compile_sig_vars(vars):
    buf = []
    for x in sorted(vars):
        if x[:3] in ('tsk', 'gen', 'bld'):
            buf.append('buf.append(%s)' % x)
    if buf:
        return funex(COMPILE_TEMPLATE_SIG_VARS % '\n\t'.join(buf))
    return None


def task_factory(
    name, func=None, vars=None, color='GREEN', ext_in=[], ext_out=[], before=[], after=[], shell=False, scan=None
):
    params = {
        'vars': vars or [],
        'color': color,
        'name': name,
        'shell': shell,
        'scan': scan,
    }
    if isinstance(func, str) or isinstance(func, tuple):
        params['run_str'] = func
    else:
        params['run'] = func
    cls = type(Task)(name, (Task,), params)
    classes[name] = cls
    if ext_in:
        cls.ext_in = Utils.to_list(ext_in)
    if ext_out:
        cls.ext_out = Utils.to_list(ext_out)
    if before:
        cls.before = Utils.to_list(before)
    if after:
        cls.after = Utils.to_list(after)
    return cls


def deep_inputs(cls):
    def sig_explicit_deps(self):
        Task.sig_explicit_deps(self)
        Task.sig_deep_inputs(self)

    cls.sig_explicit_deps = sig_explicit_deps
    return cls


TaskBase = Task


class TaskSemaphore(object):
    def __init__(self, num):
        self.num = num
        self.locking = set()
        self.waiting = set()

    def is_locked(self):
        return len(self.locking) >= self.num

    def acquire(self, tsk):
        if self.is_locked():
            raise IndexError('Cannot lock more %r' % self.locking)
        self.locking.add(tsk)

    def release(self, tsk):
        self.locking.remove(tsk)



# problematic character: é (e-acute in ISO-8859-15)
#
# NOTE: if you edit this file, be careful not to change the encoding!
#
# The first comment line in this file intentionally contains a non-ASCII character in local encoding and fails to
# declare the encoding using PEP263 encoding header.
#
# While python is able to load and run the module, retrieving its source code via loader's `get_source`
# method is expected to raise an error:
#
# ```
# import mymodule2
# mymodule2.__loader__.get_source('mymodule2')
# ```
#
# ```
# Traceback (most recent call last):
#   File "/usr/lib64/python3.11/tokenize.py", line 334, in find_cookie
#     line_string = line.decode('utf-8')
#                   ^^^^^^^^^^^^^^^^^^^^
# UnicodeDecodeError: 'utf-8' codec can't decode byte 0xe9 in position 25: invalid continuation byte
#
# During handling of the above exception, another exception occurred:
#
# Traceback (most recent call last):
#   File "<stdin>", line 1, in <module>
#   File "<frozen importlib._bootstrap_external>", line 997, in get_source
#   File "<frozen importlib._bootstrap_external>", line 768, in decode_source
#   File "/usr/lib64/python3.11/tokenize.py", line 375, in detect_encoding
#     encoding = find_cookie(first)
#                ^^^^^^^^^^^^^^^^^^
#   File "/usr/lib64/python3.11/tokenize.py", line 339, in find_cookie
#     raise SyntaxError(msg)
# SyntaxError: invalid or missing encoding declaration
# ```
#
# In contrast to `mymodule1` example, this example triggers a `SyntaxError` due to offending character appearing in the
# first line of the file, and thus breaking the scan for PEP263 encoding header.

def hello():
    return "hello"



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import pytest

from PyInstaller.utils.tests import importorskip, onedir_only
from PyInstaller.utils.hooks import can_import_module


# pywin32 provides several modules that can be imported on their own, such as win32api or win32com.
# They are located in three directories, which are added to search path via pywin32.pth:
# - win32
# - win32/lib
# - pythonwin
@importorskip('win32api')  # Use win32api to check for presence of pywin32.
@pytest.mark.parametrize(
    'module',
    (
        # Binary extensions from win32
        'mmapfile',
        'odbc',
        'perfmon',
        'servicemanager',
        'timer',
        'win32api',
        'win32clipboard',
        'win32console',
        'win32cred',
        'win32crypt',
        'win32event',
        'win32evtlog',
        'win32file',
        'win32gui',
        'win32help',
        'win32inet',
        'win32job',
        'win32lz',
        'win32net',
        'win32pdh',
        'win32pipe',
        'win32print',
        'win32process',
        'win32profile',
        'win32ras',
        'win32security',
        'win32service',
        '_win32sysloader',
        'win32trace',
        'win32transaction',
        'win32ts',
        'win32wnet',
        '_winxptheme',
        # Python modules from win32/lib
        'afxres',
        'commctrl',
        'dbi',
        'mmsystem',
        'netbios',
        'ntsecuritycon',
        'pywin32_bootstrap',
        'pywin32_testutil',
        'pywintypes',
        'rasutil',
        'regcheck',
        'regutil',
        'sspicon',
        'sspi',
        'win2kras',
        'win32con',
        'win32cryptcon',
        'win32evtlogutil',
        'win32gui_struct',
        'win32inetcon',
        'win32netcon',
        'win32pdhquery',
        'win32pdhutil',
        'win32rcparser',
        'win32serviceutil',
        'win32timezone',
        'win32traceutil',
        'win32verstamp',
        'winerror',
        'winioctlcon',
        'winnt',
        'winperf',
        'winxptheme',
        # Binary extensions from pythonwin
        'dde',
        'win32uiole',
        'win32ui',
        # Python package from pythonwin
        'pywin',
        # Packages/modules in top-level directory
        'adodbapi',
        'isapi',
        'win32com',
        'win32comext',
        'pythoncom',
    )
)
@onedir_only
def test_pywin32_imports(pyi_builder, module):
    if not can_import_module(module):
        pytest.skip(f"Module '{module}' cannot be imported.")

    # Basic import test
    pyi_builder.test_source(f"""
        import {module}
        """)


@importorskip('win32com')
def test_pywin32_win32com(pyi_builder):
    pyi_builder.test_source(
        """
        # Test importing some modules from pywin32 package.
        # All modules from pywin32 depens on module pywintypes.
        # This module should be also included.
        import win32com
        import win32com.client
        import win32com.server
        """
    )


@importorskip('win32com')
def test_pywin32_comext(pyi_builder):
    pyi_builder.test_source(
        """
        # Test importing modules from win32com that are actually present in
        # win32comext, and made available by __path__ changes in win32com.
        from win32com.shell import shell
        from win32com.propsys import propsys
        from win32com.bits import bits
        """
    )


@importorskip('win32ui')
def test_pywin32_win32ui(pyi_builder):
    pyi_builder.test_source(
        """
        # Test importing some modules from pywin32 package.
        # All modules from pywin32 depens on module pywintypes.
        # This module should be also included.
        import win32ui
        from pywin.mfc.dialog import Dialog
        d = Dialog(win32ui.IDD_SIMPLE_INPUT)
        """
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller import log as logging
from PyInstaller.utils.hooks import tcl_tk

logger = logging.getLogger(__name__)


def pre_find_module_path(hook_api):
    if not tcl_tk.tcltk_info.available:
        logger.warning("tkinter installation is broken. It will be excluded from the application")
        hook_api.search_dirs = []



#-----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# This package/module might be provided by setuptools >= 71.0.0, which makes its vendored dependencies public by
# appending path to its `setuptools._vendored` directory to `sys.path`. The following shared pre-safe-import-module
# hook implementation checks whether this is the case, and sets up aliases to prevent duplicate collection.
from PyInstaller.utils.hooks.setuptools import pre_safe_import_module  # noqa: F401



# -----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# -----------------------------------------------------------------------------

# A test package for use with `pyi_optimization.py` test program.

# Module's docstring
"""Test package."""

# Check if __debug__ is set
has_debug = __debug__

# Check if assert works
has_assert = True
try:
    assert False
    has_assert = False
except Exception:
    pass


# Function with a docstring
def test_function():
    """Test function."""
    pass



#-----------------------------------------------------------------------------
# Copyright (c) 2017-2023, PyInstaller Development Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: Apache-2.0
#-----------------------------------------------------------------------------


def _pyi_rthook():
    import sys

    import multiprocessing
    import multiprocessing.spawn

    from subprocess import _args_from_interpreter_flags

    # Prevent `spawn` from trying to read `__main__` in from the main script
    multiprocessing.process.ORIGINAL_DIR = None

    def _freeze_support():
        # We want to catch the two processes that are spawned by the multiprocessing code:
        # - the semaphore tracker, which cleans up named semaphores in the `spawn` multiprocessing mode
        # - the fork server, which keeps track of worker processes in the `forkserver` mode.
        # Both of these processes are started by spawning a new copy of the running executable, passing it the flags
        # from `_args_from_interpreter_flags` and then "-c" and an import statement.
        # Look for those flags and the import statement, then `exec()` the code ourselves.

        if (
            len(sys.argv) >= 2 and sys.argv[-2] == '-c' and sys.argv[-1].startswith(
                ('from multiprocessing.resource_tracker import main', 'from multiprocessing.forkserver import main')
            ) and set(sys.argv[1:-2]) == set(_args_from_interpreter_flags())
        ):
            exec(sys.argv[-1])
            sys.exit()

        if multiprocessing.spawn.is_forking(sys.argv):
            kwds = {}
            for arg in sys.argv[2:]:
                name, value = arg.split('=')
                if value == 'None':
                    kwds[name] = None
                else:
                    kwds[name] = int(value)
            multiprocessing.spawn.spawn_main(**kwds)
            sys.exit()

    multiprocessing.freeze_support = multiprocessing.spawn.freeze_support = _freeze_support


_pyi_rthook()
del _pyi_rthook



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re, traceback
from waflib import Utils, Logs, Errors
from waflib.Tools import fc, fc_config, fc_scan, ar, ccroot
from waflib.Configure import conf
from waflib.TaskGen import after_method, feature


@conf
def find_ifort(conf):
    fc = conf.find_program('ifort', var='FC')
    conf.get_ifort_version(fc)
    conf.env.FC_NAME = 'IFORT'


@conf
def ifort_modifier_win32(self):
    v = self.env
    v.IFORT_WIN32 = True
    v.FCSTLIB_MARKER = ''
    v.FCSHLIB_MARKER = ''
    v.FCLIB_ST = v.FCSTLIB_ST = '%s.lib'
    v.FCLIBPATH_ST = v.STLIBPATH_ST = '/LIBPATH:%s'
    v.FCINCPATH_ST = '/I%s'
    v.FCDEFINES_ST = '/D%s'
    v.fcprogram_PATTERN = v.fcprogram_test_PATTERN = '%s.exe'
    v.fcshlib_PATTERN = '%s.dll'
    v.fcstlib_PATTERN = v.implib_PATTERN = '%s.lib'
    v.FCLNK_TGT_F = '/out:'
    v.FC_TGT_F = ['/c', '/o', '']
    v.FCFLAGS_fcshlib = ''
    v.LINKFLAGS_fcshlib = '/DLL'
    v.AR_TGT_F = '/out:'
    v.IMPLIB_ST = '/IMPLIB:%s'
    v.append_value('LINKFLAGS', '/subsystem:console')
    if v.IFORT_MANIFEST:
        v.append_value('LINKFLAGS', ['/MANIFEST'])


@conf
def ifort_modifier_darwin(conf):
    fc_config.fortran_modifier_darwin(conf)


@conf
def ifort_modifier_platform(conf):
    dest_os = conf.env.DEST_OS or Utils.unversioned_sys_platform()
    ifort_modifier_func = getattr(conf, 'ifort_modifier_' + dest_os, None)
    if ifort_modifier_func:
        ifort_modifier_func()


@conf
def get_ifort_version(conf, fc):
    version_re = re.compile(r"\bIntel\b.*\bVersion\s*(?P<major>\d*)\.(?P<minor>\d*)", re.I).search
    if Utils.is_win32:
        cmd = fc
    else:
        cmd = fc + ['-logo']
    out, err = fc_config.getoutput(conf, cmd, stdin=False)
    match = version_re(out) or version_re(err)
    if not match:
        conf.fatal('cannot determine ifort version.')
    k = match.groupdict()
    conf.env.FC_VERSION = (k['major'], k['minor'])


def configure(conf):
    if Utils.is_win32:
        compiler, version, path, includes, libdirs, arch = conf.detect_ifort()
        v = conf.env
        v.DEST_CPU = arch
        v.PATH = path
        v.INCLUDES = includes
        v.LIBPATH = libdirs
        v.MSVC_COMPILER = compiler
        try:
            v.MSVC_VERSION = float(version)
        except ValueError:
            v.MSVC_VERSION = float(version[:-3])
        conf.find_ifort_win32()
        conf.ifort_modifier_win32()
    else:
        conf.find_ifort()
        conf.find_program('xiar', var='AR')
        conf.find_ar()
        conf.fc_flags()
        conf.fc_add_flags()
        conf.ifort_modifier_platform()


all_ifort_platforms = [('intel64', 'amd64'), ('em64t', 'amd64'), ('ia32', 'x86'), ('Itanium', 'ia64')]


@conf
def gather_ifort_versions(conf, versions):
    version_pattern = re.compile(r'^...?.?\....?.?')
    try:
        all_versions = Utils.winreg.OpenKey(
            Utils.winreg.HKEY_LOCAL_MACHINE, 'SOFTWARE\\Wow6432node\\Intel\\Compilers\\Fortran'
        )
    except OSError:
        try:
            all_versions = Utils.winreg.OpenKey(Utils.winreg.HKEY_LOCAL_MACHINE, 'SOFTWARE\\Intel\\Compilers\\Fortran')
        except OSError:
            return
    index = 0
    while 1:
        try:
            version = Utils.winreg.EnumKey(all_versions, index)
        except OSError:
            break
        index += 1
        if not version_pattern.match(version):
            continue
        targets = {}
        for target, arch in all_ifort_platforms:
            if target == 'intel64':
                targetDir = 'EM64T_NATIVE'
            else:
                targetDir = target
            try:
                Utils.winreg.OpenKey(all_versions, version + '\\' + targetDir)
                icl_version = Utils.winreg.OpenKey(all_versions, version)
                path, type = Utils.winreg.QueryValueEx(icl_version, 'ProductDir')
            except OSError:
                pass
            else:
                batch_file = os.path.join(path, 'bin', 'ifortvars.bat')
                if os.path.isfile(batch_file):
                    targets[target] = target_compiler(conf, 'intel', arch, version, target, batch_file)
        for target, arch in all_ifort_platforms:
            try:
                icl_version = Utils.winreg.OpenKey(all_versions, version + '\\' + target)
                path, type = Utils.winreg.QueryValueEx(icl_version, 'ProductDir')
            except OSError:
                continue
            else:
                batch_file = os.path.join(path, 'bin', 'ifortvars.bat')
                if os.path.isfile(batch_file):
                    targets[target] = target_compiler(conf, 'intel', arch, version, target, batch_file)
        major = version[0:2]
        versions['intel ' + major] = targets


@conf
def setup_ifort(conf, versiondict):
    platforms = Utils.to_list(conf.env.MSVC_TARGETS) or [i for i, j in all_ifort_platforms]
    desired_versions = conf.env.MSVC_VERSIONS or list(reversed(list(versiondict.keys())))
    for version in desired_versions:
        try:
            targets = versiondict[version]
        except KeyError:
            continue
        for arch in platforms:
            try:
                cfg = targets[arch]
            except KeyError:
                continue
            cfg.evaluate()
            if cfg.is_valid:
                compiler, revision = version.rsplit(' ', 1)
                return compiler, revision, cfg.bindirs, cfg.incdirs, cfg.libdirs, cfg.cpu
    conf.fatal(
        'ifort: Impossible to find a valid architecture for building %r - %r' %
        (desired_versions, list(versiondict.keys()))
    )


@conf
def get_ifort_version_win32(conf, compiler, version, target, vcvars):
    try:
        conf.msvc_cnt += 1
    except AttributeError:
        conf.msvc_cnt = 1
    batfile = conf.bldnode.make_node('waf-print-msvc-%d.bat' % conf.msvc_cnt)
    batfile.write(
        """@echo off
set INCLUDE=
set LIB=
call "%s" %s
echo PATH=%%PATH%%
echo INCLUDE=%%INCLUDE%%
echo LIB=%%LIB%%;%%LIBPATH%%
""" % (vcvars, target)
    )
    sout = conf.cmd_and_log(['cmd.exe', '/E:on', '/V:on', '/C', batfile.abspath()])
    batfile.delete()
    lines = sout.splitlines()
    if not lines[0]:
        lines.pop(0)
    MSVC_PATH = MSVC_INCDIR = MSVC_LIBDIR = None
    for line in lines:
        if line.startswith('PATH='):
            path = line[5:]
            MSVC_PATH = path.split(';')
        elif line.startswith('INCLUDE='):
            MSVC_INCDIR = [i for i in line[8:].split(';') if i]
        elif line.startswith('LIB='):
            MSVC_LIBDIR = [i for i in line[4:].split(';') if i]
    if None in (MSVC_PATH, MSVC_INCDIR, MSVC_LIBDIR):
        conf.fatal('ifort: Could not find a valid architecture for building (get_ifort_version_win32)')
    env = dict(os.environ)
    env.update(PATH=path)
    compiler_name, linker_name, lib_name = _get_prog_names(conf, compiler)
    fc = conf.find_program(compiler_name, path_list=MSVC_PATH)
    if 'CL' in env:
        del (env['CL'])
    try:
        conf.cmd_and_log(fc + ['/help'], env=env)
    except UnicodeError:
        st = traceback.format_exc()
        if conf.logger:
            conf.logger.error(st)
        conf.fatal('ifort: Unicode error - check the code page?')
    except Exception as e:
        Logs.debug('ifort: get_ifort_version: %r %r %r -> failure %s', compiler, version, target, str(e))
        conf.fatal('ifort: cannot run the compiler in get_ifort_version (run with -v to display errors)')
    else:
        Logs.debug('ifort: get_ifort_version: %r %r %r -> OK', compiler, version, target)
    finally:
        conf.env[compiler_name] = ''
    return (MSVC_PATH, MSVC_INCDIR, MSVC_LIBDIR)


class target_compiler(object):
    def __init__(self, ctx, compiler, cpu, version, bat_target, bat, callback=None):
        self.conf = ctx
        self.name = None
        self.is_valid = False
        self.is_done = False
        self.compiler = compiler
        self.cpu = cpu
        self.version = version
        self.bat_target = bat_target
        self.bat = bat
        self.callback = callback

    def evaluate(self):
        if self.is_done:
            return
        self.is_done = True
        try:
            vs = self.conf.get_ifort_version_win32(self.compiler, self.version, self.bat_target, self.bat)
        except Errors.ConfigurationError:
            self.is_valid = False
            return
        if self.callback:
            vs = self.callback(self, vs)
        self.is_valid = True
        (self.bindirs, self.incdirs, self.libdirs) = vs

    def __str__(self):
        return str((self.bindirs, self.incdirs, self.libdirs))

    def __repr__(self):
        return repr((self.bindirs, self.incdirs, self.libdirs))


@conf
def detect_ifort(self):
    return self.setup_ifort(self.get_ifort_versions(False))


@conf
def get_ifort_versions(self, eval_and_save=True):
    dct = {}
    self.gather_ifort_versions(dct)
    return dct


def _get_prog_names(self, compiler):
    if compiler == 'intel':
        compiler_name = 'ifort'
        linker_name = 'XILINK'
        lib_name = 'XILIB'
    else:
        compiler_name = 'CL'
        linker_name = 'LINK'
        lib_name = 'LIB'
    return compiler_name, linker_name, lib_name


@conf
def find_ifort_win32(conf):
    v = conf.env
    path = v.PATH
    compiler = v.MSVC_COMPILER
    version = v.MSVC_VERSION
    compiler_name, linker_name, lib_name = _get_prog_names(conf, compiler)
    v.IFORT_MANIFEST = (compiler == 'intel' and version >= 11)
    fc = conf.find_program(compiler_name, var='FC', path_list=path)
    env = dict(conf.environ)
    if path:
        env.update(PATH=';'.join(path))
    if not conf.cmd_and_log(fc + ['/nologo', '/help'], env=env):
        conf.fatal('not intel fortran compiler could not be identified')
    v.FC_NAME = 'IFORT'
    if not v.LINK_FC:
        conf.find_program(linker_name, var='LINK_FC', path_list=path, mandatory=True)
    if not v.AR:
        conf.find_program(lib_name, path_list=path, var='AR', mandatory=True)
        v.ARFLAGS = ['/nologo']
    if v.IFORT_MANIFEST:
        conf.find_program('MT', path_list=path, var='MT')
        v.MTFLAGS = ['/nologo']
    try:
        conf.load('winres')
    except Errors.WafError:
        Logs.warn('Resource compiler not found. Compiling resource file is disabled')


@after_method('apply_link')
@feature('fc')
def apply_flags_ifort(self):
    if not self.env.IFORT_WIN32 or not getattr(self, 'link_task', None):
        return
    is_static = isinstance(self.link_task, ccroot.stlink_task)
    subsystem = getattr(self, 'subsystem', '')
    if subsystem:
        subsystem = '/subsystem:%s' % subsystem
        flags = is_static and 'ARFLAGS' or 'LINKFLAGS'
        self.env.append_value(flags, subsystem)
    if not is_static:
        for f in self.env.LINKFLAGS:
            d = f.lower()
            if d[1:] == 'debug':
                pdbnode = self.link_task.outputs[0].change_ext('.pdb')
                self.link_task.outputs.append(pdbnode)
                if getattr(self, 'install_task', None):
                    self.pdb_install_task = self.add_install_files(
                        install_to=self.install_task.install_to, install_from=pdbnode
                    )
                break


@feature('fcprogram', 'fcshlib', 'fcprogram_test')
@after_method('apply_link')
def apply_manifest_ifort(self):
    if self.env.IFORT_WIN32 and getattr(self, 'link_task', None):
        self.link_task.env.FC = self.env.LINK_FC
    if self.env.IFORT_WIN32 and self.env.IFORT_MANIFEST and getattr(self, 'link_task', None):
        out_node = self.link_task.outputs[0]
        man_node = out_node.parent.find_or_declare(out_node.name + '.manifest')
        self.link_task.outputs.append(man_node)
        self.env.DO_MANIFEST = True



# -----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# -----------------------------------------------------------------------------

# This script checks for availability of features/flags that are turned off at higher optimization levels.

import sys
import json

# Test package that implements similar checks.
import mypackage

# Check if __debug__ is set
has_debug = __debug__

# Check if assert works
has_assert = True
try:
    assert False
    has_assert = False
except Exception:
    pass


# Function with a docstring
def test_function():
    """Test function."""
    pass


# Collect results from both entry-point script and test package.
results = {
    "sys.flags.optimize": sys.flags.optimize,
    "script": {
        "has_debug": has_debug,
        "has_assert": has_assert,
        "function_has_doc": test_function.__doc__ is not None,
    },
    "module": {
        "has_debug": mypackage.has_debug,
        "has_assert": mypackage.has_assert,
        "module_has_doc": mypackage.__doc__ is not None,
        "function_has_doc": mypackage.test_function.__doc__ is not None,
    }
}

# Write to JSON file, or dump to stdout
if len(sys.argv) > 1:
    with open(sys.argv[1], "w") as fp:
        json.dump(results, fp, indent=2)
else:
    json.dump(results, sys.stdout, indent=2)



#-----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# Test that we can start a nested `multiprocessing.Process` from within a `multiprocessing.Process`. See #7494.

import sys
import multiprocessing


def nested_process_function(queue):
    print("Running nested sub-process!")
    queue.put(2)


def process_function(queue):
    print("Running sub-process!")
    queue.put(1)

    process = multiprocessing.Process(target=nested_process_function, args=(queue,))
    process.start()
    process.join()


def main(start_method):
    multiprocessing.set_start_method(start_method)
    queue = multiprocessing.Queue()

    process = multiprocessing.Process(target=process_function, args=(queue,))
    process.start()
    process.join()

    # Read results from queue; we expect one for each process.
    # NOTE: this goes against the `multiprocessing` programming recommendations, because we should read from queue
    # before joining the feeding processes, lest we incur a deadlock when queue's buffer fills up. However, as we put in
    # only two  elements, we take a calculated risk; this way, we can read only available items post-hoc, which in turn
    # allows us to avoid blocking forever if a process happens to fail for some reason.
    results = []
    while not queue.empty():
        results.append(queue.get())

    # NOTE: as per [1]: "If multiple processes are enqueuing objects, it is possible for the objects to be received at
    # the other end out-of-order. However, objects enqueued by the same process will always be in the expected order
    # with respect to each other."
    # [1] https://docs.python.org/3/library/multiprocessing.html#pipes-and-queues
    #
    # Therefore, accept either order of results as a valid one.
    print(f"Results: {results}")
    assert results == [1, 2] or results == [2, 1]


if __name__ == '__main__':
    multiprocessing.freeze_support()

    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <start-method>")
        sys.exit(1)

    main(sys.argv[1])



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os
from waflib import Errors, Options, Task, Utils
from waflib.TaskGen import before_method, feature, extension
from waflib.Configure import conf


@feature('rubyext')
@before_method('apply_incpaths', 'process_source', 'apply_bundle', 'apply_link')
def init_rubyext(self):
    self.install_path = '${ARCHDIR_RUBY}'
    self.uselib = self.to_list(getattr(self, 'uselib', ''))
    if not 'RUBY' in self.uselib:
        self.uselib.append('RUBY')
    if not 'RUBYEXT' in self.uselib:
        self.uselib.append('RUBYEXT')


@feature('rubyext')
@before_method('apply_link', 'propagate_uselib_vars')
def apply_ruby_so_name(self):
    self.env.cshlib_PATTERN = self.env.cxxshlib_PATTERN = self.env.rubyext_PATTERN


@conf
def check_ruby_version(self, minver=()):
    ruby = self.find_program('ruby', var='RUBY', value=Options.options.rubybinary)
    try:
        version = self.cmd_and_log(ruby + ['-e', 'puts defined?(VERSION) ? VERSION : RUBY_VERSION']).strip()
    except Errors.WafError:
        self.fatal('could not determine ruby version')
    self.env.RUBY_VERSION = version
    try:
        ver = tuple(map(int, version.split('.')))
    except Errors.WafError:
        self.fatal('unsupported ruby version %r' % version)
    cver = ''
    if minver:
        cver = '> ' + '.'.join(str(x) for x in minver)
        if ver < minver:
            self.fatal('ruby is too old %r' % ver)
    self.msg('Checking for ruby version %s' % cver, version)


@conf
def check_ruby_ext_devel(self):
    if not self.env.RUBY:
        self.fatal('ruby detection is required first')
    if not self.env.CC_NAME and not self.env.CXX_NAME:
        self.fatal('load a c/c++ compiler first')
    version = tuple(map(int, self.env.RUBY_VERSION.split(".")))

    def read_out(cmd):
        return Utils.to_list(self.cmd_and_log(self.env.RUBY + ['-rrbconfig', '-e', cmd]))

    def read_config(key):
        return read_out('puts RbConfig::CONFIG[%r]' % key)

    cpppath = archdir = read_config('archdir')
    if version >= (1, 9, 0):
        ruby_hdrdir = read_config('rubyhdrdir')
        cpppath += ruby_hdrdir
        if version >= (2, 0, 0):
            cpppath += read_config('rubyarchhdrdir')
        cpppath += [os.path.join(ruby_hdrdir[0], read_config('arch')[0])]
    self.check(header_name='ruby.h', includes=cpppath, errmsg='could not find ruby header file', link_header_test=False)
    self.env.LIBPATH_RUBYEXT = read_config('libdir')
    self.env.LIBPATH_RUBYEXT += archdir
    self.env.INCLUDES_RUBYEXT = cpppath
    self.env.CFLAGS_RUBYEXT = read_config('CCDLFLAGS')
    self.env.rubyext_PATTERN = '%s.' + read_config('DLEXT')[0]
    flags = read_config('LDSHARED')
    while flags and flags[0][0] != '-':
        flags = flags[1:]
    if len(flags) > 1 and flags[1] == "ppc":
        flags = flags[2:]
    self.env.LINKFLAGS_RUBYEXT = flags
    self.env.LINKFLAGS_RUBYEXT += read_config('LIBS')
    self.env.LINKFLAGS_RUBYEXT += read_config('LIBRUBYARG_SHARED')
    if Options.options.rubyarchdir:
        self.env.ARCHDIR_RUBY = Options.options.rubyarchdir
    else:
        self.env.ARCHDIR_RUBY = read_config('sitearchdir')[0]
    if Options.options.rubylibdir:
        self.env.LIBDIR_RUBY = Options.options.rubylibdir
    else:
        self.env.LIBDIR_RUBY = read_config('sitelibdir')[0]


@conf
def check_ruby_module(self, module_name):
    self.start_msg('Ruby module %s' % module_name)
    try:
        self.cmd_and_log(self.env.RUBY + ['-e', 'require \'%s\';puts 1' % module_name])
    except Errors.WafError:
        self.end_msg(False)
        self.fatal('Could not find the ruby module %r' % module_name)
    self.end_msg(True)


@extension('.rb')
def process(self, node):
    return self.create_task('run_ruby', node)


class run_ruby(Task.Task):
    run_str = '${RUBY} ${RBFLAGS} -I ${SRC[0].parent.abspath()} ${SRC}'


def options(opt):
    opt.add_option(
        '--with-ruby-archdir',
        type='string',
        dest='rubyarchdir',
        help='Specify directory where to install arch specific files'
    )
    opt.add_option('--with-ruby-libdir', type='string', dest='rubylibdir', help='Specify alternate ruby library path')
    opt.add_option('--with-ruby-binary', type='string', dest='rubybinary', help='Specify alternate ruby binary')



# mypackage.__init__ and mypackage.a are collected into PYZ archive (the default behavior), while mypackage.b is
# collected as source .py file only.
module_collection_mode = {
    'mypackage.b': 'py',
}



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re
from waflib import Utils
from waflib.Tools import fc, fc_config, fc_scan, ar
from waflib.Configure import conf


@conf
def find_g95(conf):
    fc = conf.find_program('g95', var='FC')
    conf.get_g95_version(fc)
    conf.env.FC_NAME = 'G95'


@conf
def g95_flags(conf):
    v = conf.env
    v.FCFLAGS_fcshlib = ['-fPIC']
    v.FORTRANMODFLAG = ['-fmod=', '']
    v.FCFLAGS_DEBUG = ['-Werror']


@conf
def g95_modifier_win32(conf):
    fc_config.fortran_modifier_win32(conf)


@conf
def g95_modifier_cygwin(conf):
    fc_config.fortran_modifier_cygwin(conf)


@conf
def g95_modifier_darwin(conf):
    fc_config.fortran_modifier_darwin(conf)


@conf
def g95_modifier_platform(conf):
    dest_os = conf.env.DEST_OS or Utils.unversioned_sys_platform()
    g95_modifier_func = getattr(conf, 'g95_modifier_' + dest_os, None)
    if g95_modifier_func:
        g95_modifier_func()


@conf
def get_g95_version(conf, fc):
    version_re = re.compile(r"g95\s*(?P<major>\d*)\.(?P<minor>\d*)").search
    cmd = fc + ['--version']
    out, err = fc_config.getoutput(conf, cmd, stdin=False)
    if out:
        match = version_re(out)
    else:
        match = version_re(err)
    if not match:
        conf.fatal('cannot determine g95 version')
    k = match.groupdict()
    conf.env.FC_VERSION = (k['major'], k['minor'])


def configure(conf):
    conf.find_g95()
    conf.find_ar()
    conf.fc_flags()
    conf.fc_add_flags()
    conf.g95_flags()
    conf.g95_modifier_platform()



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
#
# This assists in creating a ``.egg`` package for use with testing ``collect_submodules``.
# To do so, execute ``python setup.py bdist_egg``.
from setuptools import setup, find_packages

setup(
    name='hookutils_egg',
    zip_safe=True,
    packages=find_packages(),
    # Manually include the fake extension modules for testing. They are not automatically included.
    package_data={'hookutils_package': ['pyextension.*']},
)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os
from waflib import Task, Options, Utils, Errors
from waflib.Configure import conf
from waflib.TaskGen import extension, feature, before_method


@before_method('apply_incpaths', 'apply_link', 'propagate_uselib_vars')
@feature('perlext')
def init_perlext(self):
    self.uselib = self.to_list(getattr(self, 'uselib', []))
    if not 'PERLEXT' in self.uselib:
        self.uselib.append('PERLEXT')
    self.env.cshlib_PATTERN = self.env.cxxshlib_PATTERN = self.env.perlext_PATTERN


@extension('.xs')
def xsubpp_file(self, node):
    outnode = node.change_ext('.c')
    self.create_task('xsubpp', node, outnode)
    self.source.append(outnode)


class xsubpp(Task.Task):
    run_str = '${PERL} ${XSUBPP} -noprototypes -typemap ${EXTUTILS_TYPEMAP} ${SRC} > ${TGT}'
    color = 'BLUE'
    ext_out = ['.h']


@conf
def check_perl_version(self, minver=None):
    res = True
    if minver:
        cver = '.'.join(map(str, minver))
    else:
        cver = ''
    self.start_msg('Checking for minimum perl version %s' % cver)
    perl = self.find_program('perl', var='PERL', value=getattr(Options.options, 'perlbinary', None))
    version = self.cmd_and_log(perl + ["-e", 'printf \"%vd\", $^V'])
    if not version:
        res = False
        version = "Unknown"
    elif not minver is None:
        ver = tuple(map(int, version.split(".")))
        if ver < minver:
            res = False
    self.end_msg(version, color=res and 'GREEN' or 'YELLOW')
    return res


@conf
def check_perl_module(self, module):
    cmd = self.env.PERL + ['-e', 'use %s' % module]
    self.start_msg('perl module %s' % module)
    try:
        r = self.cmd_and_log(cmd)
    except Errors.WafError:
        self.end_msg(False)
        return None
    self.end_msg(r or True)
    return r


@conf
def check_perl_ext_devel(self):
    env = self.env
    perl = env.PERL
    if not perl:
        self.fatal('find perl first')

    def cmd_perl_config(s):
        return perl + ['-MConfig', '-e', 'print \"%s\"' % s]

    def cfg_str(cfg):
        return self.cmd_and_log(cmd_perl_config(cfg))

    def cfg_lst(cfg):
        return Utils.to_list(cfg_str(cfg))

    def find_xsubpp():
        for var in ('privlib', 'vendorlib'):
            xsubpp = cfg_lst('$Config{%s}/ExtUtils/xsubpp$Config{exe_ext}' % var)
            if xsubpp and os.path.isfile(xsubpp[0]):
                return xsubpp
        return self.find_program('xsubpp')

    env.LINKFLAGS_PERLEXT = cfg_lst('$Config{lddlflags}')
    env.INCLUDES_PERLEXT = cfg_lst('$Config{archlib}/CORE')
    env.CFLAGS_PERLEXT = cfg_lst('$Config{ccflags} $Config{cccdlflags}')
    env.EXTUTILS_TYPEMAP = cfg_lst('$Config{privlib}/ExtUtils/typemap')
    env.XSUBPP = find_xsubpp()
    if not getattr(Options.options, 'perlarchdir', None):
        env.ARCHDIR_PERL = cfg_str('$Config{sitearch}')
    else:
        env.ARCHDIR_PERL = getattr(Options.options, 'perlarchdir')
    env.perlext_PATTERN = '%s.' + cfg_str('$Config{dlext}')


def options(opt):
    opt.add_option(
        '--with-perl-binary', type='string', dest='perlbinary', help='Specify alternate perl binary', default=None
    )
    opt.add_option(
        '--with-perl-archdir',
        type='string',
        dest='perlarchdir',
        help='Specify directory where to install arch specific files',
        default=None
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

name = 'pyi_testmod_relimp.pyi_testmod_relimp.relimp2'

from . import relimp3  # noqa: E402
assert relimp3.name == 'pyi_testmod_relimp.pyi_testmod_relimp.relimp3'

from .. import pyi_testmod_relimp  # noqa: E402
assert pyi_testmod_relimp.name == 'pyi_testmod_relimp.pyi_testmod_relimp'

import pyi_testmod_relimp  # noqa: E402
assert pyi_testmod_relimp.name == 'pyi_testmod_relimp'

import pyi_testmod_relimp.relimp2  # noqa: E402
assert pyi_testmod_relimp.relimp2.name == 'pyi_testmod_relimp.relimp2'

# While this seems to work when running Python, it is wrong:
#  .pyi_testmod_relimp should be a sibling of this package
#from .pyi_testmod_relimp import relimp2  # noqa: E402
#assert pyi_testmod_relimp2.name == 'pyi_testmod_relimp.relimp2'



#-----------------------------------------------------------------------------
# Copyright (c) 2025, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import collections
import os
import io
import marshal
import zipfile

from PyInstaller.loader.pyimod01_archive import PYZ_ITEM_MODULE, PYZ_ITEM_PKG
from PyInstaller.archive.readers import CArchiveReader, PKG_ITEM_PYPACKAGE, PKG_ITEM_PYMODULE, PKG_ITEM_PYSOURCE


def _analyze_collected_code_objects(filename):
    # Construct lists of expected and found co_filename values, along with metadata (pkg/pyz and name). We compare these
    # lists at the end, and rely on pytest to produce a nicely formatted diff if there is a mismatch....
    Entry = collections.namedtuple('Entry', ['location', 'name', 'co_filename'])
    expected_entries = []
    found_entries = []

    # PKG archive: all modules and scripts - bootstrap modules, bootstrap script, run-time hooks, entry-point script.
    pkg_archive = CArchiveReader(filename)

    for name, (*_, typecode) in pkg_archive.toc.items():
        if typecode not in {PKG_ITEM_PYPACKAGE, PKG_ITEM_PYMODULE, PKG_ITEM_PYSOURCE}:
            continue

        expected_co_filename = name + '.py'
        expected_entries.append(Entry('PKG', name, expected_co_filename))

        code_object = marshal.loads(pkg_archive.extract(name))
        found_entries.append(Entry('PKG', name, code_object.co_filename))

    # base_library.zip - either in PKG archive, or in _internal next to executable.
    baselib_zip_filename = os.path.join(os.path.dirname(filename), '_internal', 'base_library.zip')
    if os.path.isfile(baselib_zip_filename):
        baselib_zip = zipfile.ZipFile(baselib_zip_filename, mode='r')
    else:
        # onefile mode
        baselib_zip = zipfile.ZipFile(io.BytesIO(pkg_archive.extract('base_library.zip')), mode='r')

    for entry in baselib_zip.infolist():
        # Turn the entry filename into expected co_filename by changing .pyc extension into .py. The entries in .zip
        # file have POSIX separators, which need to be normalized for comparison on Windows.
        expected_co_filename = os.path.normpath(entry.filename)[:-1]
        expected_entries.append(Entry('base_library.zip', entry.filename, expected_co_filename))

        code_object = marshal.loads(baselib_zip.read(entry)[16:])  # skip the .pyc header (16 bytes)
        found_entries.append(Entry('base_library.zip', entry.filename, code_object.co_filename))

    # PYZ archive: all modules and packages.
    pyz_archive = pkg_archive.open_embedded_archive('PYZ.pyz')

    for name, (typecode, *_) in pyz_archive.toc.items():
        # Analyze only modules and packages (i.e., ignore PEP-420 namespace package entries).
        if typecode not in {PYZ_ITEM_MODULE, PYZ_ITEM_PKG}:
            continue

        if typecode == PYZ_ITEM_MODULE:
            expected_co_filename = os.path.join(*name.split('.')) + '.py'
        else:
            expected_co_filename = os.path.join(*name.split('.'), '__init__.py')
        expected_entries.append(Entry('PYZ', name, expected_co_filename))

        code_object = pyz_archive.extract(name)
        found_entries.append(Entry('PYZ', name, code_object.co_filename))

    # Compare the obtained lists
    assert expected_entries == found_entries


def test_co_filename_anonymization(pyi_builder):
    pyi_args = [
        # stdlib packages with run-time hooks
        '--hiddenimport',
        'inspect',
        '--hiddenimport',
        'multiprocessing',
        # setuptools for its vendored packages
        '--hiddenimport',
        'setuptools',
        # PyInstaller's "fake" modules/packages
        '--hiddenimport',
        'pyi_splash',
        '--hiddenimport',
        '_pyi_rth_utils.tempfile',
        '--hiddenimport',
        '_pyi_rth_utils.qt',
    ]

    pyi_builder.test_source("""
        print("Hello!")
        """, pyi_args=pyi_args)

    executables = pyi_builder._find_executables("test_source")
    assert len(executables) == 1

    _analyze_collected_code_objects(executables[0])



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re, sys
from waflib import Utils

wlock = Utils.threading.Lock()
try:
    from ctypes import Structure, windll, c_short, c_ushort, c_ulong, c_int, byref, c_wchar, POINTER, c_long
except ImportError:

    class AnsiTerm(object):
        def __init__(self, stream):
            self.stream = stream
            try:
                self.errors = self.stream.errors
            except AttributeError:
                pass
            self.encoding = self.stream.encoding

        def write(self, txt):
            try:
                wlock.acquire()
                self.stream.write(txt)
                self.stream.flush()
            finally:
                wlock.release()

        def fileno(self):
            return self.stream.fileno()

        def flush(self):
            self.stream.flush()

        def isatty(self):
            return self.stream.isatty()
else:

    class COORD(Structure):
        _fields_ = [("X", c_short), ("Y", c_short)]

    class SMALL_RECT(Structure):
        _fields_ = [("Left", c_short), ("Top", c_short), ("Right", c_short), ("Bottom", c_short)]

    class CONSOLE_SCREEN_BUFFER_INFO(Structure):
        _fields_ = [("Size", COORD), ("CursorPosition", COORD), ("Attributes", c_ushort), ("Window", SMALL_RECT),
                    ("MaximumWindowSize", COORD)]

    class CONSOLE_CURSOR_INFO(Structure):
        _fields_ = [('dwSize', c_ulong), ('bVisible', c_int)]

    try:
        _type = unicode
    except NameError:
        _type = str
    to_int = lambda number, default: number and int(number) or default
    STD_OUTPUT_HANDLE = -11
    STD_ERROR_HANDLE = -12
    windll.kernel32.GetStdHandle.argtypes = [c_ulong]
    windll.kernel32.GetStdHandle.restype = c_ulong
    windll.kernel32.GetConsoleScreenBufferInfo.argtypes = [c_ulong, POINTER(CONSOLE_SCREEN_BUFFER_INFO)]
    windll.kernel32.GetConsoleScreenBufferInfo.restype = c_long
    windll.kernel32.SetConsoleTextAttribute.argtypes = [c_ulong, c_ushort]
    windll.kernel32.SetConsoleTextAttribute.restype = c_long
    windll.kernel32.FillConsoleOutputCharacterW.argtypes = [c_ulong, c_wchar, c_ulong, POINTER(COORD), POINTER(c_ulong)]
    windll.kernel32.FillConsoleOutputCharacterW.restype = c_long
    windll.kernel32.FillConsoleOutputAttribute.argtypes = [c_ulong, c_ushort, c_ulong, POINTER(COORD), POINTER(c_ulong)]
    windll.kernel32.FillConsoleOutputAttribute.restype = c_long
    windll.kernel32.SetConsoleCursorPosition.argtypes = [c_ulong, POINTER(COORD)]
    windll.kernel32.SetConsoleCursorPosition.restype = c_long
    windll.kernel32.SetConsoleCursorInfo.argtypes = [c_ulong, POINTER(CONSOLE_CURSOR_INFO)]
    windll.kernel32.SetConsoleCursorInfo.restype = c_long

    class AnsiTerm(object):
        def __init__(self, s):
            self.stream = s
            try:
                self.errors = s.errors
            except AttributeError:
                pass
            self.encoding = s.encoding
            self.cursor_history = []
            handle = (s.fileno() == 2) and STD_ERROR_HANDLE or STD_OUTPUT_HANDLE
            self.hconsole = windll.kernel32.GetStdHandle(handle)
            self._sbinfo = CONSOLE_SCREEN_BUFFER_INFO()
            self._csinfo = CONSOLE_CURSOR_INFO()
            windll.kernel32.GetConsoleCursorInfo(self.hconsole, byref(self._csinfo))
            self._orig_sbinfo = CONSOLE_SCREEN_BUFFER_INFO()
            r = windll.kernel32.GetConsoleScreenBufferInfo(self.hconsole, byref(self._orig_sbinfo))
            self._isatty = r == 1

        def screen_buffer_info(self):
            windll.kernel32.GetConsoleScreenBufferInfo(self.hconsole, byref(self._sbinfo))
            return self._sbinfo

        def clear_line(self, param):
            mode = param and int(param) or 0
            sbinfo = self.screen_buffer_info()
            if mode == 1:
                line_start = COORD(0, sbinfo.CursorPosition.Y)
                line_length = sbinfo.Size.X
            elif mode == 2:
                line_start = COORD(sbinfo.CursorPosition.X, sbinfo.CursorPosition.Y)
                line_length = sbinfo.Size.X - sbinfo.CursorPosition.X
            else:
                line_start = sbinfo.CursorPosition
                line_length = sbinfo.Size.X - sbinfo.CursorPosition.X
            chars_written = c_ulong()
            windll.kernel32.FillConsoleOutputCharacterW(
                self.hconsole, c_wchar(' '), line_length, line_start, byref(chars_written)
            )
            windll.kernel32.FillConsoleOutputAttribute(
                self.hconsole, sbinfo.Attributes, line_length, line_start, byref(chars_written)
            )

        def clear_screen(self, param):
            mode = to_int(param, 0)
            sbinfo = self.screen_buffer_info()
            if mode == 1:
                clear_start = COORD(0, 0)
                clear_length = sbinfo.CursorPosition.X * sbinfo.CursorPosition.Y
            elif mode == 2:
                clear_start = COORD(0, 0)
                clear_length = sbinfo.Size.X * sbinfo.Size.Y
                windll.kernel32.SetConsoleCursorPosition(self.hconsole, clear_start)
            else:
                clear_start = sbinfo.CursorPosition
                clear_length = ((sbinfo.Size.X - sbinfo.CursorPosition.X) + sbinfo.Size.X *
                                (sbinfo.Size.Y - sbinfo.CursorPosition.Y))
            chars_written = c_ulong()
            windll.kernel32.FillConsoleOutputCharacterW(
                self.hconsole, c_wchar(' '), clear_length, clear_start, byref(chars_written)
            )
            windll.kernel32.FillConsoleOutputAttribute(
                self.hconsole, sbinfo.Attributes, clear_length, clear_start, byref(chars_written)
            )

        def push_cursor(self, param):
            sbinfo = self.screen_buffer_info()
            self.cursor_history.append(sbinfo.CursorPosition)

        def pop_cursor(self, param):
            if self.cursor_history:
                old_pos = self.cursor_history.pop()
                windll.kernel32.SetConsoleCursorPosition(self.hconsole, old_pos)

        def set_cursor(self, param):
            y, sep, x = param.partition(';')
            x = to_int(x, 1) - 1
            y = to_int(y, 1) - 1
            sbinfo = self.screen_buffer_info()
            new_pos = COORD(min(max(0, x), sbinfo.Size.X), min(max(0, y), sbinfo.Size.Y))
            windll.kernel32.SetConsoleCursorPosition(self.hconsole, new_pos)

        def set_column(self, param):
            x = to_int(param, 1) - 1
            sbinfo = self.screen_buffer_info()
            new_pos = COORD(min(max(0, x), sbinfo.Size.X), sbinfo.CursorPosition.Y)
            windll.kernel32.SetConsoleCursorPosition(self.hconsole, new_pos)

        def move_cursor(self, x_offset=0, y_offset=0):
            sbinfo = self.screen_buffer_info()
            new_pos = COORD(
                min(max(0, sbinfo.CursorPosition.X + x_offset), sbinfo.Size.X),
                min(max(0, sbinfo.CursorPosition.Y + y_offset), sbinfo.Size.Y)
            )
            windll.kernel32.SetConsoleCursorPosition(self.hconsole, new_pos)

        def move_up(self, param):
            self.move_cursor(y_offset=-to_int(param, 1))

        def move_down(self, param):
            self.move_cursor(y_offset=to_int(param, 1))

        def move_left(self, param):
            self.move_cursor(x_offset=-to_int(param, 1))

        def move_right(self, param):
            self.move_cursor(x_offset=to_int(param, 1))

        def next_line(self, param):
            sbinfo = self.screen_buffer_info()
            self.move_cursor(x_offset=-sbinfo.CursorPosition.X, y_offset=to_int(param, 1))

        def prev_line(self, param):
            sbinfo = self.screen_buffer_info()
            self.move_cursor(x_offset=-sbinfo.CursorPosition.X, y_offset=-to_int(param, 1))

        def rgb2bgr(self, c):
            return ((c & 1) << 2) | (c & 2) | ((c & 4) >> 2)

        def set_color(self, param):
            cols = param.split(';')
            sbinfo = self.screen_buffer_info()
            attr = sbinfo.Attributes
            for c in cols:
                c = to_int(c, 0)
                if 29 < c < 38:
                    attr = (attr & 0xfff0) | self.rgb2bgr(c - 30)
                elif 39 < c < 48:
                    attr = (attr & 0xff0f) | (self.rgb2bgr(c - 40) << 4)
                elif c == 0:
                    attr = self._orig_sbinfo.Attributes
                elif c == 1:
                    attr |= 0x08
                elif c == 4:
                    attr |= 0x80
                elif c == 7:
                    attr = (attr & 0xff88) | ((attr & 0x70) >> 4) | ((attr & 0x07) << 4)
            windll.kernel32.SetConsoleTextAttribute(self.hconsole, attr)

        def show_cursor(self, param):
            self._csinfo.bVisible = 1
            windll.kernel32.SetConsoleCursorInfo(self.hconsole, byref(self._csinfo))

        def hide_cursor(self, param):
            self._csinfo.bVisible = 0
            windll.kernel32.SetConsoleCursorInfo(self.hconsole, byref(self._csinfo))

        ansi_command_table = {
            'A': move_up,
            'B': move_down,
            'C': move_right,
            'D': move_left,
            'E': next_line,
            'F': prev_line,
            'G': set_column,
            'H': set_cursor,
            'f': set_cursor,
            'J': clear_screen,
            'K': clear_line,
            'h': show_cursor,
            'l': hide_cursor,
            'm': set_color,
            's': push_cursor,
            'u': pop_cursor,
        }
        ansi_tokens = re.compile(r'(?:\x1b\[([0-9?;]*)([a-zA-Z])|([^\x1b]+))')

        def write(self, text):
            try:
                wlock.acquire()
                if self._isatty:
                    for param, cmd, txt in self.ansi_tokens.findall(text):
                        if cmd:
                            cmd_func = self.ansi_command_table.get(cmd)
                            if cmd_func:
                                cmd_func(self, param)
                        else:
                            self.writeconsole(txt)
                else:
                    self.stream.write(text)
            finally:
                wlock.release()

        def writeconsole(self, txt):
            chars_written = c_ulong()
            writeconsole = windll.kernel32.WriteConsoleA
            if isinstance(txt, _type):
                writeconsole = windll.kernel32.WriteConsoleW
            done = 0
            todo = len(txt)
            chunk = 32 << 10
            while todo != 0:
                doing = min(chunk, todo)
                buf = txt[done:done + doing]
                r = writeconsole(self.hconsole, buf, doing, byref(chars_written), None)
                if r == 0:
                    chunk >>= 1
                    continue
                done += doing
                todo -= doing

        def fileno(self):
            return self.stream.fileno()

        def flush(self):
            pass

        def isatty(self):
            return self._isatty

    if sys.stdout.isatty() or sys.stderr.isatty():
        handle = sys.stdout.isatty() and STD_OUTPUT_HANDLE or STD_ERROR_HANDLE
        console = windll.kernel32.GetStdHandle(handle)
        sbinfo = CONSOLE_SCREEN_BUFFER_INFO()

        def get_term_cols():
            windll.kernel32.GetConsoleScreenBufferInfo(console, byref(sbinfo))
            return sbinfo.Size.X - 1


try:
    import struct, fcntl, termios
except ImportError:
    pass
else:
    if (sys.stdout.isatty() or sys.stderr.isatty()) and os.environ.get('TERM', '') not in ('dumb', 'emacs'):
        FD = sys.stdout.isatty() and sys.stdout.fileno() or sys.stderr.fileno()

        def fun():
            return struct.unpack("HHHH", fcntl.ioctl(FD, termios.TIOCGWINSZ, struct.pack("HHHH", 0, 0, 0, 0)))[1]

        try:
            fun()
        except Exception as e:
            pass
        else:
            get_term_cols = fun



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os
import functools
from waflib import Context, Task, Utils, Options, Errors, Logs
from waflib.TaskGen import taskgen_method, before_method, feature, extension
from waflib.Configure import conf


@taskgen_method
def add_marshal_file(self, filename, prefix):
    if not hasattr(self, 'marshal_list'):
        self.marshal_list = []
    self.meths.append('process_marshal')
    self.marshal_list.append((filename, prefix))


@before_method('process_source')
def process_marshal(self):
    for f, prefix in getattr(self, 'marshal_list', []):
        node = self.path.find_resource(f)
        if not node:
            raise Errors.WafError('file not found %r' % f)
        h_node = node.change_ext('.h')
        c_node = node.change_ext('.c')
        task = self.create_task('glib_genmarshal', node, [h_node, c_node])
        task.env.GLIB_GENMARSHAL_PREFIX = prefix
    self.source = self.to_nodes(getattr(self, 'source', []))
    self.source.append(c_node)


class glib_genmarshal(Task.Task):
    vars = ['GLIB_GENMARSHAL_PREFIX', 'GLIB_GENMARSHAL']
    color = 'BLUE'
    ext_out = ['.h']

    def run(self):
        bld = self.generator.bld
        get = self.env.get_flat
        cmd1 = "%s %s --prefix=%s --header > %s" % (
            get('GLIB_GENMARSHAL'), self.inputs[0].srcpath(), get('GLIB_GENMARSHAL_PREFIX'), self.outputs[0].abspath()
        )
        ret = bld.exec_command(cmd1)
        if ret:
            return ret
        c = '''#include "%s"\n''' % self.outputs[0].name
        self.outputs[1].write(c)
        cmd2 = "%s %s --prefix=%s --body >> %s" % (
            get('GLIB_GENMARSHAL'), self.inputs[0].srcpath(), get('GLIB_GENMARSHAL_PREFIX'), self.outputs[1].abspath()
        )
        return bld.exec_command(cmd2)


@taskgen_method
def add_enums_from_template(self, source='', target='', template='', comments=''):
    if not hasattr(self, 'enums_list'):
        self.enums_list = []
    self.meths.append('process_enums')
    self.enums_list.append({
        'source': source,
        'target': target,
        'template': template,
        'file-head': '',
        'file-prod': '',
        'file-tail': '',
        'enum-prod': '',
        'value-head': '',
        'value-prod': '',
        'value-tail': '',
        'comments': comments
    })


@taskgen_method
def add_enums(
    self,
    source='',
    target='',
    file_head='',
    file_prod='',
    file_tail='',
    enum_prod='',
    value_head='',
    value_prod='',
    value_tail='',
    comments=''
):
    if not hasattr(self, 'enums_list'):
        self.enums_list = []
    self.meths.append('process_enums')
    self.enums_list.append({
        'source': source,
        'template': '',
        'target': target,
        'file-head': file_head,
        'file-prod': file_prod,
        'file-tail': file_tail,
        'enum-prod': enum_prod,
        'value-head': value_head,
        'value-prod': value_prod,
        'value-tail': value_tail,
        'comments': comments
    })


@before_method('process_source')
def process_enums(self):
    for enum in getattr(self, 'enums_list', []):
        task = self.create_task('glib_mkenums')
        env = task.env
        inputs = []
        source_list = self.to_list(enum['source'])
        if not source_list:
            raise Errors.WafError('missing source ' + str(enum))
        source_list = [self.path.find_resource(k) for k in source_list]
        inputs += source_list
        env.GLIB_MKENUMS_SOURCE = [k.abspath() for k in source_list]
        if not enum['target']:
            raise Errors.WafError('missing target ' + str(enum))
        tgt_node = self.path.find_or_declare(enum['target'])
        if tgt_node.name.endswith('.c'):
            self.source.append(tgt_node)
        env.GLIB_MKENUMS_TARGET = tgt_node.abspath()
        options = []
        if enum['template']:
            template_node = self.path.find_resource(enum['template'])
            options.append('--template %s' % (template_node.abspath()))
            inputs.append(template_node)
        params = {
            'file-head': '--fhead',
            'file-prod': '--fprod',
            'file-tail': '--ftail',
            'enum-prod': '--eprod',
            'value-head': '--vhead',
            'value-prod': '--vprod',
            'value-tail': '--vtail',
            'comments': '--comments'
        }
        for param, option in params.items():
            if enum[param]:
                options.append('%s %r' % (option, enum[param]))
        env.GLIB_MKENUMS_OPTIONS = ' '.join(options)
        task.set_inputs(inputs)
        task.set_outputs(tgt_node)


class glib_mkenums(Task.Task):
    run_str = '${GLIB_MKENUMS} ${GLIB_MKENUMS_OPTIONS} ${GLIB_MKENUMS_SOURCE} > ${GLIB_MKENUMS_TARGET}'
    color = 'PINK'
    ext_out = ['.h']


@taskgen_method
def add_settings_schemas(self, filename_list):
    if not hasattr(self, 'settings_schema_files'):
        self.settings_schema_files = []
    if not isinstance(filename_list, list):
        filename_list = [filename_list]
    self.settings_schema_files.extend(filename_list)


@taskgen_method
def add_settings_enums(self, namespace, filename_list):
    if hasattr(self, 'settings_enum_namespace'):
        raise Errors.WafError("Tried to add gsettings enums to %r more than once" % self.name)
    self.settings_enum_namespace = namespace
    if not isinstance(filename_list, list):
        filename_list = [filename_list]
    self.settings_enum_files = filename_list


@feature('glib2')
def process_settings(self):
    enums_tgt_node = []
    install_files = []
    settings_schema_files = getattr(self, 'settings_schema_files', [])
    if settings_schema_files and not self.env.GLIB_COMPILE_SCHEMAS:
        raise Errors.WafError(
            "Unable to process GSettings schemas - glib-compile-schemas was not found during configure"
        )
    if hasattr(self, 'settings_enum_files'):
        enums_task = self.create_task('glib_mkenums')
        source_list = self.settings_enum_files
        source_list = [self.path.find_resource(k) for k in source_list]
        enums_task.set_inputs(source_list)
        enums_task.env.GLIB_MKENUMS_SOURCE = [k.abspath() for k in source_list]
        target = self.settings_enum_namespace + '.enums.xml'
        tgt_node = self.path.find_or_declare(target)
        enums_task.set_outputs(tgt_node)
        enums_task.env.GLIB_MKENUMS_TARGET = tgt_node.abspath()
        enums_tgt_node = [tgt_node]
        install_files.append(tgt_node)
        options = '--comments "<!-- @comment@ -->" --fhead "<schemalist>" --vhead "  <@type@ id=\\"%s.@EnumName@\\">" --vprod "    <value nick=\\"@valuenick@\\" value=\\"@valuenum@\\"/>" --vtail "  </@type@>" --ftail "</schemalist>" ' % (
            self.settings_enum_namespace
        )
        enums_task.env.GLIB_MKENUMS_OPTIONS = options
    for schema in settings_schema_files:
        schema_task = self.create_task('glib_validate_schema')
        schema_node = self.path.find_resource(schema)
        if not schema_node:
            raise Errors.WafError("Cannot find the schema file %r" % schema)
        install_files.append(schema_node)
        source_list = enums_tgt_node + [schema_node]
        schema_task.set_inputs(source_list)
        schema_task.env.GLIB_COMPILE_SCHEMAS_OPTIONS = [("--schema-file=" + k.abspath()) for k in source_list]
        target_node = schema_node.change_ext('.xml.valid')
        schema_task.set_outputs(target_node)
        schema_task.env.GLIB_VALIDATE_SCHEMA_OUTPUT = target_node.abspath()

    def compile_schemas_callback(bld):
        if not bld.is_install:
            return
        compile_schemas = Utils.to_list(bld.env.GLIB_COMPILE_SCHEMAS)
        destdir = Options.options.destdir
        paths = bld._compile_schemas_registered
        if destdir:
            paths = (os.path.join(destdir, path.lstrip(os.sep)) for path in paths)
        for path in paths:
            Logs.pprint('YELLOW', 'Updating GSettings schema cache %r' % path)
            if self.bld.exec_command(compile_schemas + [path]):
                Logs.warn('Could not update GSettings schema cache %r' % path)

    if self.bld.is_install:
        schemadir = self.env.GSETTINGSSCHEMADIR
        if not schemadir:
            raise Errors.WafError(
                'GSETTINGSSCHEMADIR not defined (should have been set up automatically during configure)'
            )
        if install_files:
            self.add_install_files(install_to=schemadir, install_from=install_files)
            registered_schemas = getattr(self.bld, '_compile_schemas_registered', None)
            if not registered_schemas:
                registered_schemas = self.bld._compile_schemas_registered = set()
                self.bld.add_post_fun(compile_schemas_callback)
            registered_schemas.add(schemadir)


class glib_validate_schema(Task.Task):
    run_str = 'rm -f ${GLIB_VALIDATE_SCHEMA_OUTPUT} && ${GLIB_COMPILE_SCHEMAS} --dry-run ${GLIB_COMPILE_SCHEMAS_OPTIONS} && touch ${GLIB_VALIDATE_SCHEMA_OUTPUT}'
    color = 'PINK'


@extension('.gresource.xml')
def process_gresource_source(self, node):
    if not self.env.GLIB_COMPILE_RESOURCES:
        raise Errors.WafError(
            "Unable to process GResource file - glib-compile-resources was not found during configure"
        )
    if 'gresource' in self.features:
        return
    h_node = node.change_ext('_xml.h')
    c_node = node.change_ext('_xml.c')
    self.create_task('glib_gresource_source', node, [h_node, c_node])
    self.source.append(c_node)


@feature('gresource')
def process_gresource_bundle(self):
    for i in self.to_list(self.source):
        node = self.path.find_resource(i)
        task = self.create_task('glib_gresource_bundle', node, node.change_ext(''))
        inst_to = getattr(self, 'install_path', None)
        if inst_to:
            self.add_install_files(install_to=inst_to, install_from=task.outputs)


class glib_gresource_base(Task.Task):
    color = 'BLUE'
    base_cmd = '${GLIB_COMPILE_RESOURCES} --sourcedir=${SRC[0].parent.srcpath()} --sourcedir=${SRC[0].bld_dir()}'

    def scan(self):
        bld = self.generator.bld
        kw = {}
        kw['cwd'] = self.get_cwd()
        kw['quiet'] = Context.BOTH
        cmd = Utils.subst_vars(
            '${GLIB_COMPILE_RESOURCES} --sourcedir=%s --sourcedir=%s --generate-dependencies %s' %
            (self.inputs[0].parent.srcpath(), self.inputs[0].bld_dir(), self.inputs[0].bldpath()), self.env
        )
        output = bld.cmd_and_log(cmd, **kw)
        nodes = []
        names = []
        for dep in output.splitlines():
            if dep:
                node = bld.bldnode.find_node(dep)
                if node:
                    nodes.append(node)
                else:
                    names.append(dep)
        return (nodes, names)


class glib_gresource_source(glib_gresource_base):
    vars = ['GLIB_COMPILE_RESOURCES']
    fun_h = Task.compile_fun_shell(
        glib_gresource_base.base_cmd + ' --target=${TGT[0].abspath()} --generate-header ${SRC}'
    )
    fun_c = Task.compile_fun_shell(
        glib_gresource_base.base_cmd + ' --target=${TGT[1].abspath()} --generate-source ${SRC}'
    )
    ext_out = ['.h']

    def run(self):
        return self.fun_h[0](self) or self.fun_c[0](self)


class glib_gresource_bundle(glib_gresource_base):
    run_str = glib_gresource_base.base_cmd + ' --target=${TGT} ${SRC}'
    shell = True


@conf
def find_glib_genmarshal(conf):
    conf.find_program('glib-genmarshal', var='GLIB_GENMARSHAL')


@conf
def find_glib_mkenums(conf):
    if not conf.env.PERL:
        conf.find_program('perl', var='PERL')
    conf.find_program('glib-mkenums', interpreter='PERL', var='GLIB_MKENUMS')


@conf
def find_glib_compile_schemas(conf):
    conf.find_program('glib-compile-schemas', var='GLIB_COMPILE_SCHEMAS')

    def getstr(varname):
        return getattr(Options.options, varname, getattr(conf.env, varname, ''))

    gsettingsschemadir = getstr('GSETTINGSSCHEMADIR')
    if not gsettingsschemadir:
        datadir = getstr('DATADIR')
        if not datadir:
            prefix = conf.env.PREFIX
            datadir = os.path.join(prefix, 'share')
        gsettingsschemadir = os.path.join(datadir, 'glib-2.0', 'schemas')
    conf.env.GSETTINGSSCHEMADIR = gsettingsschemadir


@conf
def find_glib_compile_resources(conf):
    conf.find_program('glib-compile-resources', var='GLIB_COMPILE_RESOURCES')


def configure(conf):
    conf.find_glib_genmarshal()
    conf.find_glib_mkenums()
    conf.find_glib_compile_schemas(mandatory=False)
    conf.find_glib_compile_resources(mandatory=False)


def options(opt):
    gr = opt.add_option_group('Installation directories')
    gr.add_option(
        '--gsettingsschemadir',
        help='GSettings schema location [DATADIR/glib-2.0/schemas]',
        default='',
        dest='GSETTINGSSCHEMADIR'
    )



import basemod

import inspect
import sys


class _Popen(basemod.Popen):
    def __init__(self, *args, **kw):
        print(inspect.getfile(self.__init__))
        print(inspect.getfile(super().__init__))
        super().__init__(*args, **kw)


# Reduce recursion limit to shorten the traceback.
sys.setrecursionlimit(50)

basemod.Popen = _Popen
p = basemod.Popen()



#-----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import sys
import multiprocessing


def process_function():
    import dis
    import sys

    main_code = sys.modules["__main__"].__loader__.get_code("__main__")
    print(dis.dis(main_code))


def main(start_method):
    # Set start method
    multiprocessing.set_start_method(start_method)

    # Start a sub-process
    process = multiprocessing.Process(target=process_function)
    process.start()
    process.join()

    # Ensure process finished successfully
    assert process.exitcode == 0, f"Process exited with non-success code {process.exitcode}!"


if __name__ == '__main__':
    multiprocessing.freeze_support()

    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <start-method>")
        sys.exit(1)

    main(sys.argv[1])



"""
Tests that deal with pep420 namespace packages.

PEP 420 is new in Python 3.3
"""
import os
import shutil
import sys
import subprocess
import textwrap
import unittest

from PyInstaller.lib.modulegraph import modulegraph

gRootDir = os.path.dirname(os.path.abspath(__file__))
gSrcDir = os.path.join(gRootDir, 'testpkg-pep420-namespace')

class TestPythonBehaviour (unittest.TestCase):
    def importModule(self, name):
        test_dir1 = os.path.join(gSrcDir, 'path1')
        test_dir2 = os.path.join(gSrcDir, 'path2')
        if '.' in name:
            script = textwrap.dedent("""\
                import site
                site.addsitedir(%r)
                site.addsitedir(%r)
                try:
                    import %s
                except ImportError:
                    import %s
                print (%s.__name__)
            """) %(test_dir1, test_dir2, name, name.rsplit('.', 1)[0], name)
        else:
            script = textwrap.dedent("""\
                import site
                site.addsitedir(%r)
                site.addsitedir(%r)
                import %s
                print (%s.__name__)
            """) %(test_dir1, test_dir2, name, name)

        p = subprocess.Popen([sys.executable, '-c', script],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                cwd=os.path.join(
                    os.path.dirname(os.path.abspath(__file__)),
                    'testpkg-relimport'),
                encoding='utf8',
        )
        data = p.communicate()[0]
        data = data.strip()
        if data.endswith(' refs]'):
            data = data.rsplit('\n', 1)[0].strip()

        sts = p.wait()

        if sts != 0:
            print (data)
            self.fail("import of %r failed"%(name,))

        return data

    def testToplevel(self):
        m = self.importModule('package.sub1')
        self.assertEqual(m, 'package.sub1')

        m = self.importModule('package.sub2')
        self.assertEqual(m, 'package.sub2')

    def testSub(self):
        m = self.importModule('package.subpackage.sub')
        self.assertEqual(m, 'package.subpackage.sub')

        m = self.importModule('package.nspkg.mod')
        self.assertEqual(m, 'package.nspkg.mod')

class TestModuleGraphImport (unittest.TestCase):
    if not hasattr(unittest.TestCase, 'assertIsInstance'):
        def assertIsInstance(self, value, types):
            if not isinstance(value, types):
                self.fail("%r is not an instance of %r", value, types)

    def setUp(self):
        self.mf = modulegraph.ModuleGraph(path=[
                os.path.join(gSrcDir, 'path1'),
                os.path.join(gSrcDir, 'path2'),
            ] + sys.path)


    def testRootPkg(self):
        self.mf.import_hook('package')

        node = self.mf.find_node('package')
        self.assertIsInstance(node, modulegraph.NamespacePackage)
        self.assertEqual(node.identifier, 'package')
        self.assertEqual(node.filename, '-')

    def testRootPkgModule(self):
        self.mf.import_hook('package.sub1')

        node = self.mf.find_node('package.sub1')
        self.assertIsInstance(node, modulegraph.SourceModule)
        self.assertEqual(node.identifier, 'package.sub1')

        self.mf.import_hook('package.sub2')
        node = self.mf.find_node('package.sub2')
        self.assertIsInstance(node, modulegraph.SourceModule)
        self.assertEqual(node.identifier, 'package.sub2')

    def testSubRootPkgModule(self):
        self.mf.import_hook('package.subpackage.sub')

        node = self.mf.find_node('package.subpackage.sub')
        self.assertIsInstance(node, modulegraph.SourceModule)
        self.assertEqual(node.identifier, 'package.subpackage.sub')

        node = self.mf.find_node('package')
        self.assertIsInstance(node, modulegraph.NamespacePackage)

        self.mf.import_hook('package.nspkg.mod')
        node = self.mf.find_node('package.nspkg.mod')
        self.assertIsInstance(node, modulegraph.SourceModule)
        self.assertEqual(node.identifier, 'package.nspkg.mod')


if __name__ == "__main__":
    unittest.main()



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------


class ExecCommandFailed(SystemExit):
    pass


class HookError(Exception):
    """
    Base class for hook related errors.
    """
    pass


class ImportErrorWhenRunningHook(HookError):
    def __str__(self):
        return (
            "ERROR: Failed to import module {0} required by hook for module {1}. Please check whether module {0} "
            "actually exists and whether the hook is compatible with your version of {1}: You might want to read more "
            "about hooks in the manual and provide a pull-request to improve PyInstaller.".format(
                self.args[0], self.args[1]
            )
        )


class RemovedCipherFeatureError(SystemExit):
    def __init__(self, message):
        super().__init__(
            f"ERROR: Bytecode encryption was removed in PyInstaller v6.0. {message}"
            " For the rationale and alternatives see https://github.com/pyinstaller/pyinstaller/pull/6999"
        )


class RemovedExternalManifestError(SystemExit):
    def __init__(self, message):
        super().__init__(f"ERROR: Support for external executable manifest was removed in PyInstaller v6.0. {message}")


class RemovedWinSideBySideSupportError(SystemExit):
    def __init__(self, message):
        super().__init__(
            f"ERROR: Support for collecting and processing WinSxS assemblies was removed in PyInstaller v6.0. {message}"
        )


class PythonLibraryNotFoundError(SystemExit):
    def __init__(self, message):
        super().__init__(f"ERROR: {message}")


class InvalidSrcDestTupleError(SystemExit):
    def __init__(self, src_dest, message):
        super().__init__(f"ERROR: Invalid (SRC, DEST_DIR) tuple: {src_dest!r}. {message}")


class ImportlibMetadataError(SystemExit):
    def __init__(self):
        super().__init__(
            "ERROR: PyInstaller requires importlib.metadata from python >= 3.10 stdlib or importlib_metadata from "
            "importlib-metadata >= 4.6"
        )



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Automatically build a spec file containing the description of the project.
"""

import argparse
import os

import PyInstaller.building.makespec
import PyInstaller.log

try:
    from argcomplete import autocomplete
except ImportError:

    def autocomplete(parser):
        return None


def generate_parser():
    p = argparse.ArgumentParser()
    PyInstaller.building.makespec.__add_options(p)
    PyInstaller.log.__add_options(p)
    p.add_argument(
        'scriptname',
        nargs='+',
    )
    return p


def run():
    p = generate_parser()
    autocomplete(p)
    args = p.parse_args()
    PyInstaller.log.__process_options(p, args)

    # Split pathex by using the path separator.
    temppaths = args.pathex[:]
    args.pathex = []
    for p in temppaths:
        args.pathex.extend(p.split(os.pathsep))

    try:
        name = PyInstaller.building.makespec.main(args.scriptname, **vars(args))
        print('Wrote %s.' % name)
        print('Now run pyinstaller.py to build the executable.')
    except KeyboardInterrupt:
        raise SystemExit("Aborted by user request.")


if __name__ == '__main__':
    run()



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Logging module for PyInstaller.
"""

__all__ = ['getLogger', 'INFO', 'WARN', 'DEBUG', 'TRACE', 'ERROR', 'FATAL', 'DEPRECATION']

import os
import logging
from logging import DEBUG, ERROR, FATAL, INFO, WARN, getLogger

TRACE = DEBUG - 5
logging.addLevelName(TRACE, 'TRACE')
DEPRECATION = WARN + 5
logging.addLevelName(DEPRECATION, 'DEPRECATION')
LEVELS = {
    'TRACE': TRACE,
    'DEBUG': DEBUG,
    'INFO': INFO,
    'WARN': WARN,
    'DEPRECATION': DEPRECATION,
    'ERROR': ERROR,
    'FATAL': FATAL,
}

FORMAT = '%(relativeCreated)d %(levelname)s: %(message)s'
_env_level = os.environ.get("PYI_LOG_LEVEL", "INFO")
try:
    level = LEVELS[_env_level.upper()]
except KeyError:
    raise SystemExit(f"ERROR: Invalid PYI_LOG_LEVEL value '{_env_level}'. Should be one of {list(LEVELS)}.")
logging.basicConfig(format=FORMAT, level=level)
logger = getLogger('PyInstaller')


def __add_options(parser):
    parser.add_argument(
        '--log-level',
        choices=LEVELS,
        metavar="LEVEL",
        dest='loglevel',
        help='Amount of detail in build-time console messages. LEVEL may be one of %s (default: INFO). '
        'Also settable via and overrides the PYI_LOG_LEVEL environment variable.' % ', '.join(LEVELS),
    )


def __process_options(parser, opts):
    if opts.loglevel:
        try:
            level = opts.loglevel.upper()
            _level = LEVELS[level]
        except KeyError:
            parser.error('Unknown log level `%s`' % opts.loglevel)
        logger.setLevel(_level)
        os.environ["PYI_LOG_LEVEL"] = level
